# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections.abc
import numbers
from typing import Optional, Any, Sequence, Mapping, Tuple, Union, Iterable

import pyproj

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonDateSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema


class InputConfig(JsonObject):
    def __init__(self,
                 store_id: str = None,
                 opener_id: str = None,
                 data_id: str = None,
                 store_params: Mapping[str, Any] = None,
                 open_params: Mapping[str, Any] = None):
        assert_true(store_id or opener_id,
                    'One of store_id and opener_id must be given')
        assert_given(data_id, 'data_id')
        self.store_id = store_id
        self.opener_id = opener_id
        self.data_id = data_id
        self.store_params = store_params
        self.open_params = open_params

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                store_id=JsonStringSchema(min_length=1),
                opener_id=JsonStringSchema(min_length=1),
                data_id=JsonStringSchema(min_length=1),
                store_params=JsonObjectSchema(additional_properties=True,
                                              nullable=True),
                open_params=JsonObjectSchema(additional_properties=True,
                                             nullable=True)
            ),
            additional_properties=False,
            required=['data_id'],
            factory=cls,
        )


class CallbackConfig(JsonObject):
    def __init__(self,
                 api_uri: str = None,
                 access_token: str = None):
        assert_true(api_uri and access_token,
                    'Both, api_uri and access_token must be given')
        self.api_uri = api_uri
        self.access_token = access_token

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                api_uri=JsonStringSchema(min_length=1),
                access_token=JsonStringSchema(min_length=1)
            ),
            additional_properties=False,
            required=["api_uri", "access_token"],
            factory=cls,
        )


class OutputConfig(JsonObject):

    def __init__(self,
                 store_id: str = None,
                 writer_id: str = None,
                 data_id: str = None,
                 store_params: Mapping[str, Any] = None,
                 write_params: Mapping[str, Any] = None,
                 replace: bool = None):
        assert_true(store_id or writer_id,
                    'One of store_id and writer_id must be given')
        self.store_id = store_id
        self.writer_id = writer_id
        self.data_id = data_id
        self.store_params = store_params
        self.write_params = write_params
        self.replace = replace

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                store_id=JsonStringSchema(min_length=1),
                writer_id=JsonStringSchema(min_length=1),
                data_id=JsonStringSchema(default=None),
                store_params=JsonObjectSchema(additional_properties=True,
                                              nullable=True),
                write_params=JsonObjectSchema(additional_properties=True,
                                              nullable=True),
                replace=JsonBooleanSchema(default=False),
            ),
            additional_properties=False,
            required=[],
            factory=cls,
        )


# Need to be aligned with params in resample_cube(cube, **params)
class CubeConfig(JsonObject):

    def __init__(self,
                 variable_names: Sequence[str] = None,
                 crs: str = None,
                 bbox: Tuple[float, float, float, float] = None,
                 spatial_res: Union[float, Tuple[float]] = None,
                 tile_size: Union[int, Tuple[int, int]] = None,
                 time_range: Tuple[str, Optional[str]] = None,
                 time_period: str = None,
                 chunks: Mapping[str, Optional[int]] = None,
                 metadata: Mapping[str, Any] = None,
                 variable_metadata: Mapping[str, Mapping[str, Any]] = None,):

        self.variable_names = None
        if variable_names is not None:
            assert_true(len(variable_names) > 0, 'variable_names is invalid')
            self.variable_names = tuple(map(str, variable_names))

        self.crs = None
        if crs is not None:
            assert_instance(crs, str, 'crs')
            try:
                pyproj.crs.CRS.from_string(crs)
            except pyproj.exceptions.CRSError:
                raise ValueError('crs is invalid')
            self.crs = crs

        self.bbox = None
        if bbox is not None:
            assert_true(len(bbox) == 4, 'bbox is invalid')
            self.bbox = tuple(map(float, bbox))

        self.spatial_res = None
        if spatial_res is not None:
            assert_instance(spatial_res, numbers.Number, 'spatial_res')
            self.spatial_res = float(spatial_res)

        self.tile_size = None
        if tile_size is not None:
            if isinstance(tile_size, int):
                tile_width, tile_height = tile_size, tile_size
            else:
                try:
                    tile_width, tile_height = tile_size
                except (ValueError, TypeError):
                    raise ValueError('tile_size must be an integer '
                                     'or a pair of integers')
                assert_instance(tile_width, numbers.Number,
                                'tile_width of tile_size')
                assert_instance(tile_height, numbers.Number,
                                'tile_height of tile_size')
            self.tile_size = tile_width, tile_height

        self.time_range = None
        if time_range is not None:
            assert_true(len(time_range) == 2, 'time_range is invalid')
            self.time_range = tuple(time_range)

        self.time_period = None
        if time_period is not None:
            assert_instance(time_period, str, 'time_period')
            self.time_period = time_period

        self.chunks = None
        if chunks is not None:
            assert_instance(chunks, collections.abc.Mapping, 'chunks')
            for chunk_dim, chunk_size in chunks.items():
                assert_instance(chunk_dim, str, 'chunk dimension')
                assert_instance(chunk_size, (int, type(None)), 'chunk size')
            self.chunks = dict(chunks)

        self.metadata = None
        if metadata is not None:
            assert_instance(metadata,
                            collections.abc.Mapping, 'metadata')
            self.metadata = dict(metadata)

        self.variable_metadata = None
        if variable_metadata is not None:
            assert_instance(variable_metadata,
                            collections.abc.Mapping, 'variable_metadata')
            for var_name, metadata in variable_metadata.items():
                assert_instance(var_name, str, 'chunk dimension')
                assert_instance(metadata,
                                collections.abc.Mapping, 'variable metadata')
            self.variable_metadata = dict(variable_metadata)

    def drop_props(self, name: Union[str, Iterable[str]]):
        """
        Drop one or more named properties from this configuration.
        :param name: the name of the property to be dropped.
        :param names: more names of properties to be dropped.
        :return: a new cube configuration.
        """
        if not name:
            return self
        name_set = {name} if isinstance(name, str) else set(name)
        for k in name_set:
            assert_true(hasattr(self, k),
                        message=f'{k} is not a property of {CubeConfig!r}')
        return self.from_dict({k: v
                               for k, v in self.to_dict().items()
                               if k not in name_set})

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                variable_names=JsonArraySchema(
                    nullable=True,
                    items=JsonStringSchema(min_length=1),
                    min_items=0
                ),
                crs=JsonStringSchema(
                    nullable=True,
                    min_length=1
                ),
                bbox=JsonArraySchema(
                    nullable=True,
                    items=[JsonNumberSchema(),
                           JsonNumberSchema(),
                           JsonNumberSchema(),
                           JsonNumberSchema()]
                ),
                spatial_res=JsonNumberSchema(
                    nullable=True,
                    exclusive_minimum=0.0),
                tile_size=JsonArraySchema(
                    nullable=True,
                    items=[
                        JsonIntegerSchema(minimum=1, maximum=2500),
                        JsonIntegerSchema(minimum=1, maximum=2500),
                    ]),
                time_range=JsonDateSchema.new_range(nullable=True),
                time_period=JsonStringSchema(
                    nullable=True,
                    pattern=r'^([1-9][0-9]*)?[DWMY]$'
                ),
                chunks=JsonObjectSchema(
                    nullable=True,
                    additional_properties=JsonIntegerSchema(nullable=True,
                                                            minimum=1)
                ),
                metadata=JsonObjectSchema(
                    nullable=True,
                    additional_properties=True
                ),
                variable_metadata=JsonObjectSchema(
                    nullable=True,
                    additional_properties=JsonObjectSchema(
                        additional_properties=True
                    )
                ),
            ),
            additional_properties=False,
            factory=cls
        )



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Tuple, Sequence, Dict, Optional, Mapping, Union, Hashable

import numpy as np
import xarray as xr

from xcube.core.gridmapping import GridMapping


class CubeSchema:
    """
    A schema that can be used to create new xcube datasets.
    The given *shape*, *dims*, and *chunks*, *coords* apply to all data variables.

    :param shape: A tuple of dimension sizes.
    :param coords: A dictionary of coordinate variables. Must have values for all *dims*.
    :param dims: A sequence of dimension names. Defaults to ``('time', 'lat', 'lon')``.
    :param chunks: A tuple of chunk sizes in each dimension.
    """

    def __init__(self,
                 shape: Sequence[int],
                 coords: Mapping[str, xr.DataArray],
                 x_name: str = 'lon',
                 y_name: str = 'lat',
                 time_name: str = 'time',
                 dims: Sequence[str] = None,
                 chunks: Sequence[int] = None):

        if not shape:
            raise ValueError('shape must be a sequence of integer sizes')
        if not coords:
            raise ValueError('coords must be a mapping from dimension names to label arrays')
        if not x_name:
            raise ValueError('x_name must be given')
        if not y_name:
            raise ValueError('y_name must be given')
        if not time_name:
            raise ValueError('time_name must be given')

        ndim = len(shape)
        if ndim < 3:
            raise ValueError('shape must have at least three dimensions')
        dims = tuple(dims) or (time_name, y_name, x_name)
        if dims and len(dims) != ndim:
            raise ValueError('dims must have same length as shape')
        if x_name not in coords or y_name not in coords or time_name not in coords:
            raise ValueError(f'missing variables {x_name!r}, {y_name!r}, {time_name!r} in coords')
        x_var, y_var, time_var = coords.get(x_name), coords.get(y_name), coords.get(time_name)
        if x_var.ndim != 1 or y_var.ndim != 1 or time_var.ndim != 1:
            raise ValueError(f'variables {x_name!r}, {y_name!r}, {time_name!r} in coords must be 1-D')
        x_dim, y_dim, time_dim = x_var.dims[0], y_var.dims[0], time_var.dims[0]
        if dims[0] != time_dim:
            raise ValueError(f"the first dimension in dims must be {time_dim!r}")
        if dims[-2:] != (y_dim, x_dim):
            raise ValueError(f"the last two dimensions in dims must be {y_dim!r} and {x_dim!r}")
        if chunks and len(chunks) != ndim:
            raise ValueError('chunks must have same length as shape')
        for i in range(ndim):
            dim_name = dims[i]
            dim_size = shape[i]
            if dim_name not in coords:
                raise ValueError(f'missing dimension {dim_name!r} in coords')
            dim_labels = coords[dim_name]
            if len(dim_labels.shape) != 1:
                raise ValueError(f'labels of {dim_name!r} in coords must be one-dimensional')
            if len(dim_labels) != dim_size:
                raise ValueError(f'number of labels of {dim_name!r} in coords does not match shape')

        self._shape = tuple(shape)
        self._x_name = x_name
        self._y_name = y_name
        self._time_name = time_name
        self._dims = dims
        self._chunks = tuple(chunks) if chunks else None
        self._coords = dict(coords)

    @property
    def ndim(self) -> int:
        """Number of dimensions."""
        return len(self._dims)

    @property
    def dims(self) -> Tuple[str, ...]:
        """Tuple of dimension names."""
        return self._dims

    @property
    def x_name(self) -> str:
        """Name of the spatial x coordinate variable."""
        return self._x_name

    @property
    def y_name(self) -> str:
        """Name of the spatial y coordinate variable."""
        return self._y_name

    @property
    def time_name(self) -> str:
        """Name of the time coordinate variable."""
        return self._time_name

    @property
    def x_var(self) -> xr.DataArray:
        """Spatial x coordinate variable."""
        return self._coords[self._x_name]

    @property
    def y_var(self) -> xr.DataArray:
        """Spatial y coordinate variable."""
        return self._coords[self._y_name]

    @property
    def time_var(self) -> xr.DataArray:
        """Time coordinate variable."""
        return self._coords[self._time_name]

    @property
    def x_dim(self) -> str:
        """Name of the spatial x dimension."""
        return self._dims[-1]

    @property
    def y_dim(self) -> str:
        """Name of the spatial y dimension."""
        return self._dims[-2]

    @property
    def time_dim(self) -> str:
        """Name of the time dimension."""
        return self._dims[0]

    @property
    def x_size(self) -> int:
        """Size of the spatial x dimension."""
        return self._shape[-1]

    @property
    def y_size(self) -> int:
        """Size of the spatial y dimension."""
        return self._shape[-2]

    @property
    def time_size(self) -> int:
        """Size of the time dimension."""
        return self._shape[0]

    @property
    def shape(self) -> Tuple[int, ...]:
        """Tuple of dimension sizes."""
        return self._shape

    @property
    def chunks(self) -> Optional[Tuple[int]]:
        """Tuple of dimension chunk sizes."""
        return self._chunks

    @property
    def coords(self) -> Dict[str, xr.DataArray]:
        """Dictionary of coordinate variables."""
        return self._coords

    @classmethod
    def new(cls, cube: xr.Dataset) -> 'CubeSchema':
        """Create a cube schema from given *cube*."""
        return get_cube_schema(cube)

    def _repr_html_(self):
        """Return a HTML representation for Jupyter Notebooks."""
        return (
            f'<table>'
            f'<tr><td>Shape:</td><td>{self.shape}</td></tr>'
            f'<tr><td>Chunk sizes:</td><td>{self.chunks}</td></tr>'
            f'<tr><td>Dimensions:</td><td>{self.dims}</td></tr>'
            f'</table>'
        )


# TODO (forman): code duplication with xcube.core.verify._check_data_variables(), line 76
def get_cube_schema(cube: xr.Dataset) -> CubeSchema:
    """
    Derive cube schema from given *cube*.

    :param cube: The data cube.
    :return: The cube schema.
    """

    xy_var_names = get_dataset_xy_var_names(cube, must_exist=True, dataset_arg_name='cube')
    time_var_name = get_dataset_time_var_name(cube, must_exist=True, dataset_arg_name='cube')

    first_dims = None
    first_shape = None
    first_chunks = None
    first_coords = None

    for var_name, var in cube.data_vars.items():

        dims = var.dims
        if first_dims is None:
            first_dims = dims
        elif first_dims != dims:
            raise ValueError(f'all variables must have same dimensions, but variable {var_name!r} '
                             f'has dimensions {dims!r}')

        shape = var.shape
        if first_shape is None:
            first_shape = shape
        elif first_shape != shape:
            raise ValueError(f'all variables must have same shape, but variable {var_name!r} '
                             f'has shape {shape!r}')

        coords = var.coords
        if first_coords is None:
            first_coords = coords

        dask_chunks = var.chunks
        if dask_chunks:
            chunks = []
            for i in range(var.ndim):
                dim_name = var.dims[i]
                dim_chunk_sizes = dask_chunks[i]
                first_size = dim_chunk_sizes[0]
                if any(size != first_size for size in dim_chunk_sizes[1:-1]):
                    raise ValueError(f'dimension {dim_name!r} of variable {var_name!r} has chunks of different sizes: '
                                     f'{dim_chunk_sizes!r}')
                chunks.append(first_size)
            chunks = tuple(chunks)
            if first_chunks is None:
                first_chunks = chunks
            elif first_chunks != chunks:
                raise ValueError(f'all variables must have same chunks, but variable {var_name!r} '
                                 f'has chunks {chunks!r}')

    if first_dims is None:
        raise ValueError('cube is empty')

    return CubeSchema(first_shape,
                      first_coords,
                      x_name=xy_var_names[0],
                      y_name=xy_var_names[1],
                      time_name=time_var_name,
                      dims=tuple(str(d) for d in first_dims),
                      chunks=first_chunks)


def get_dataset_xy_var_names(coords: Union[xr.Dataset,
                                           xr.DataArray,
                                           Mapping[Hashable, xr.DataArray]],
                             must_exist: bool = False,
                             dataset_arg_name: str = 'dataset') \
        -> Optional[Tuple[str, str]]:
    if hasattr(coords, 'coords'):
        coords = coords.coords
    x_var_name = None
    y_var_name = None
    for var_name, var in coords.items():
        if var.attrs.get('standard_name') == 'projection_x_coordinate' \
                or var.attrs.get('long_name') == 'x coordinate of projection':
            if var.ndim == 1:
                x_var_name = var_name
        if var.attrs.get('standard_name') == 'projection_y_coordinate' \
                or var.attrs.get('long_name') == 'y coordinate of projection':
            if var.ndim == 1:
                y_var_name = var_name
        if x_var_name and y_var_name:
            return str(x_var_name), str(y_var_name)

    x_var_name = None
    y_var_name = None
    for var_name, var in coords.items():
        if var.attrs.get('long_name') == 'longitude':
            if var.ndim == 1:
                x_var_name = var_name
        if var.attrs.get('long_name') == 'latitude':
            if var.ndim == 1:
                y_var_name = var_name
        if x_var_name and y_var_name:
            return str(x_var_name), str(y_var_name)

    for x_var_name, y_var_name in (('lon', 'lat'),
                                   ('longitude', 'latitude'),
                                   ('x', 'y')):
        if x_var_name in coords and y_var_name in coords:
            x_var = coords[x_var_name]
            y_var = coords[y_var_name]
            if x_var.ndim == 1 and y_var.ndim == 1:
                return x_var_name, y_var_name

    if must_exist:
        raise ValueError(f'{dataset_arg_name}'
                         f' has no valid spatial coordinate variables')


def get_dataset_time_var_name(dataset: Union[xr.Dataset, xr.DataArray],
                              must_exist: bool = False,
                              dataset_arg_name: str = 'dataset') -> Optional[str]:
    time_var_name = 'time'
    if time_var_name in dataset.coords:
        time_var = dataset.coords[time_var_name]
        if time_var.ndim == 1 and np.issubdtype(time_var.dtype, np.datetime64):
            return time_var_name

    if must_exist:
        raise ValueError(f'{dataset_arg_name} has no valid time coordinate variable')

    return None


def get_dataset_bounds_var_name(dataset: Union[xr.Dataset, xr.DataArray],
                                var_name: str,
                                must_exist: bool = False,
                                dataset_arg_name: str = 'dataset') -> Optional[str]:
    if var_name in dataset.coords:
        var = dataset[var_name]
        bounds_var_name = var.attrs.get('bounds', f'{var_name}_bnds')
        if bounds_var_name in dataset:
            bounds_var = dataset[bounds_var_name]
            if bounds_var.ndim == 2 \
                    and bounds_var.shape[0] == var.shape[0] and bounds_var.shape[1] == 2:
                return bounds_var_name

    if must_exist:
        raise ValueError(f'{dataset_arg_name} has no valid bounds variable for variable {var_name!r}')

    return None


def get_dataset_chunks(dataset: xr.Dataset) -> Dict[Hashable, int]:
    """
    Get the most common chunk sizes for each
    chunked dimension of *dataset*.

    Note: Only data variables are considered.

    :param dataset: A dataset.
    :return: A dictionary that maps dimension names
        to common chunk sizes.
    """

    # Record the frequencies of chunk sizes for
    # each dimension d in each data variable var
    dim_size_counts: Dict[Hashable, Dict[int, int]] = {}
    for var_name, var in dataset.data_vars.items():
        if var.chunks:
            for d, c in zip(var.dims, var.chunks):
                # compute max chunk size max_c from
                # e.g.  c = (512, 512, 512, 193)
                max_c = max(0, *c)
                # for dimension d, save the frequencies
                # of the different max_c
                if d not in dim_size_counts:
                    size_counts = {max_c: 1}
                    dim_size_counts[d] = size_counts
                else:
                    size_counts = dim_size_counts[d]
                    if max_c not in size_counts:
                        size_counts[max_c] = 1
                    else:
                        size_counts[max_c] += 1

    # For each dimension d, determine the most frequently
    # seen chunk size max_c
    dim_sizes: Dict[Hashable, int] = {}
    for d, size_counts in dim_size_counts.items():
        max_count = 0
        best_max_c = 0
        for max_c, count in size_counts.items():
            if count > max_count:
                # Should always come here, because count=1 is minimum
                max_count = count
                best_max_c = max_c
        assert best_max_c > 0
        dim_sizes[d] = best_max_c

    return dim_sizes


def rechunk_cube(cube: xr.Dataset,
                 gm: GridMapping,
                 chunks: Optional[Dict[str, int]] = None,
                 tile_size: Optional[Tuple[int, int]] = None) \
        -> Tuple[xr.Dataset, GridMapping]:
    """
    Re-chunk data variables of *cube* so they all share the same chunk
    sizes for their dimensions.

    This functions rechunks *cube* for maximum compatibility with
    the Zarr format. Therefore it removes the "chunks" encoding
    from all variables.

    :param cube: A data cube
    :param gm: The cube's grid mapping
    :param chunks: Optional mapping of dimension names to chunk sizes
    :param tile_size: Optional tile sizes, i.e. chunk size of
        spatial dimensions, given as (width, height)
    :return: A potentially rechunked *cube* and adjusted grid mapping.
    """

    # get initial, common cube chunk sizes from given cube
    cube_chunks = get_dataset_chunks(cube)

    # Given chunks will overwrite initial values
    if chunks:
        for dim_name, size in chunks.items():
            cube_chunks[dim_name] = size

    # Given tile size will overwrite spatial dims
    x_dim_name, y_dim_name = gm.xy_dim_names
    if tile_size is not None:
        cube_chunks[x_dim_name] = tile_size[0]
        cube_chunks[y_dim_name] = tile_size[1]

    # Given grid mapping's tile size will overwrite
    # spatial dims only if missing still
    if gm.tile_size is not None:
        if x_dim_name not in cube_chunks:
            cube_chunks[x_dim_name] = gm.tile_size[0]
        if y_dim_name not in cube_chunks:
            cube_chunks[y_dim_name] = gm.tile_size[1]

    # If there is no chunking required, return identities
    if not cube_chunks:
        return cube, gm

    chunked_cube = xr.Dataset(attrs=cube.attrs)

    # Coordinate variables are always
    # chunked automatically
    chunked_cube = chunked_cube.assign_coords(
        coords={
            var_name: var.chunk({
                dim_name: 'auto'
                for dim_name in var.dims
            })
            for var_name, var in cube.coords.items()
        }
    )

    # Data variables are chunked according to cube_chunks,
    # or if not specified, by the dimension size.
    chunked_cube = chunked_cube.assign(
        variables={
            var_name: var.chunk({
                dim_name: cube_chunks.get(dim_name, cube.dims[dim_name])
                for dim_name in var.dims
            })
            for var_name, var in cube.data_vars.items()
        }
    )

    # Update chunks encoding for Zarr
    for var_name, var in chunked_cube.variables.items():
        if 'chunks' in var.encoding:
            del var.encoding['chunks']
        # if var.chunks is not None:
        #     # sizes[0] is the first of
        #     # e.g. sizes = (512, 512, 71)
        #     var.encoding.update(chunks=[
        #         sizes[0] for sizes in var.chunks
        #     ])
        # elif 'chunks' in var.encoding:
        #     del var.encoding['chunks']
        # print(f"--> {var_name}: encoding={var.encoding.get('chunks')!r}, chunks={var.chunks!r}")

    # Test whether tile size has changed after re-chunking.
    # If so, we will change the grid mapping too.
    tile_width = cube_chunks.get(x_dim_name)
    tile_height = cube_chunks.get(y_dim_name)
    assert tile_width is not None
    assert tile_height is not None
    tile_size = (tile_width, tile_height)
    if tile_size != gm.tile_size:
        # Note, changing grid mapping tile size may
        # rechunk (2D) coordinates in chunked_cube too
        gm = gm.derive(tile_size=tile_size)

    return chunked_cube, gm



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import warnings
from typing import Dict, Any, MutableMapping

import click

from xcube.cli.common import (cli_option_quiet,
                              cli_option_verbosity,
                              cli_option_dry_run,
                              configure_cli_output)
from xcube.constants import LOG

DELETE_ATTR_VALUE = '__delete__'


# noinspection PyShadowingBuiltins
@click.command(name='patch')
@click.argument('DATASET')
@click.option('--metadata', 'metadata_path', metavar='METADATA',
              help='The metadata to be patched.'
                   ' Must be a JSON or YAML file'
                   ' using Zarr consolidated metadata format.')
@click.option('--options', 'options_path', metavar='OPTIONS',
              help='Protocol-specific storage options (see fsspec).'
                   ' Must be a JSON or YAML file.')
@cli_option_quiet
@cli_option_verbosity
@cli_option_dry_run
def patch(dataset,
          metadata_path,
          options_path,
          quiet,
          verbosity,
          dry_run):
    """
    Patch and consolidate the metadata of a dataset.

    DATASET can be either a local filesystem path or a URL.
    It must point to either a Zarr dataset (*.zarr)
    or a xcube multi-level dataset (*.levels).
    Additional storage options for a given protocol may be passed
    by the OPTIONS option.

    In METADATA, the special attribute value "__delete__" can be used to
    remove that attribute from dataset or array metadata.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)
    metadata = load_metadata(metadata_path)
    storage_options = load_storage_options(options_path)
    patch_dataset(dataset, storage_options, metadata, dry_run)


def load_metadata(metadata_path):
    from xcube.util.config import load_json_or_yaml_config

    if not metadata_path:
        raise click.ClickException('Missing metadata to be patched')

    metadata = load_json_or_yaml_config(metadata_path)
    return parse_metadata(metadata)


def parse_metadata(metadata):
    if not isinstance(metadata, dict) \
            or 'zarr_consolidated_format' not in metadata:
        raise click.ClickException(
            'Invalid consolidated metadata format'
        )
    zarr_consolidated_format = metadata.get('zarr_consolidated_format')
    if zarr_consolidated_format != 1:
        raise click.ClickException(
            'Unsupported consolidated metadata version'
        )
    metadata = metadata.get('metadata')
    if not isinstance(metadata, dict):
        raise click.ClickException('Invalid metadata format')
    _metadata = dict()
    for k, v in metadata.items():
        if not isinstance(v, dict):
            raise click.ClickException(f'Invalid metadata format:'
                                       f' entry "{k}" is not an object')
        parts = k.split('/')
        if parts[-1] != '.zattrs':
            warnings.warn(f'Ignoring metadata entry "{k}":'
                          f' can only patch "*/.zattrs"')
        elif len(parts) not in (1, 2):
            warnings.warn(f'Ignoring metadata entry {k}:'
                          f' can only patch ".zattrs" of first level')
        else:
            _metadata[k] = v
    if not _metadata:
        raise click.ClickException('No metadata provided')
    return _metadata


def load_storage_options(options_path):
    from xcube.util.config import load_json_or_yaml_config
    storage_options = {}
    if options_path:
        storage_options = load_json_or_yaml_config(options_path)
        if not isinstance(storage_options, dict):
            raise click.ClickException('Invalid storage options format')
    return storage_options


def patch_dataset(dataset_path: str,
                  storage_options: Dict[str, Any],
                  metadata: Dict[str, Any],
                  dry_run: bool):
    if dataset_path.endswith('.levels'):
        fs, root = get_fs_and_root(dataset_path, storage_options)
        protocol = fs.protocol \
            if isinstance(fs.protocol, str) \
            else fs.protocol[0]
        prefix = f"{protocol}://" if protocol != "file" else ""
        for item in fs.listdir(root, detail=False):
            if item.endswith('.zarr'):
                _patch_dataset(prefix + item,
                               storage_options, metadata, dry_run)
    else:
        _patch_dataset(dataset_path,
                       storage_options, metadata, dry_run)


def _patch_dataset(dataset_path: str,
                   storage_options: Dict[str, Any],
                   metadata: Dict[str, Any],
                   dry_run: bool):
    import zarr
    LOG.info(f'Opening {dataset_path}...')
    zarr_store = _open_zarr_store(dataset_path, storage_options)
    group = zarr.open(zarr_store, mode='r+' if not dry_run else 'r')
    for k, v in metadata.items():
        parts = k.split('/')
        item = None
        if len(parts) == 1:
            item = group
        else:
            item_name, _ = parts
            if item_name in group:
                item = group[item_name]
            else:
                warnings.warn(f'Ignoring metadata entry {k}:'
                              f' not found in dataset')
        if item is not None:
            upd_count = 0
            del_count = 0
            for ak, av in v.items():
                if av == DELETE_ATTR_VALUE:
                    if ak in item.attrs:
                        if not dry_run:
                            del item.attrs[ak]
                        del_count += 1
                else:
                    if not dry_run:
                        item.attrs[ak] = av
                    upd_count += 1
            if upd_count and del_count:
                LOG.info(f'{k}:'
                         f' {upd_count} attribute(s) updated,'
                         f' {del_count} deleted')
            elif upd_count:
                LOG.info(f'{k}: {upd_count} attribute(s) updated')
            elif del_count:
                LOG.info(f'{k}: {del_count} attribute(s) deleted')

    if not dry_run:
        zarr.convenience.consolidate_metadata(zarr_store)
    LOG.info(f'Consolidated {dataset_path}')


def _open_zarr_store(dataset_path: str, storage_options: Dict[str, Any]) \
        -> MutableMapping[str, bytes]:
    fs, root = get_fs_and_root(dataset_path, storage_options)
    return fs.get_mapper(root=root)


def get_fs_and_root(dataset_path: str, storage_options: Dict[str, Any]):
    import fsspec
    protocol, root = fsspec.core.split_protocol(dataset_path)
    protocol = protocol or 'file'
    fs = fsspec.filesystem(protocol, **storage_options)
    return fs, root



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Dict, Any, Callable

import jsonschema

from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonObjectSchema
from .error import DataStoreError


def assert_valid_params(params: Optional[Dict[str, Any]],
                        schema: Optional[JsonObjectSchema] = None,
                        name: str = 'params'):
    """
    Assert that constructor/method parameters *params* are valid.

    :param params: Dictionary of keyword arguments
        passed to a constructor/method.
    :param schema: The JSON Schema that *params* must adhere to.
    :param name: Name of the *params* variable.
    """
    _assert_valid(params, schema,
                  name, 'parameterization',
                  _validate_params)


def assert_valid_config(config: Optional[Dict[str, Any]],
                        schema: Optional[JsonObjectSchema] = None,
                        name: str = 'config'):
    """
    Assert that JSON object *config* is valid.

    :param config: JSON object used for configuration.
    :param schema: The JSON Schema that *config* must adhere to.
    :param name: Name of the *config* variable.
    """
    _assert_valid(config, schema,
                  name, 'configuration',
                  _validate_config)


def _validate_params(params: Dict[str, Any], schema: JsonObjectSchema):
    # Note, params is a dictionary of Python objects.
    # Convert them to a JSON instance
    # and perform JSON Schema validation
    schema.validate_instance(params)


def _validate_config(config: Dict[str, Any], schema: JsonObjectSchema):
    # Note, config is already a JSON object.
    # Perform JSON Schema validation directly.
    schema.validate_instance(config)


def _assert_valid(obj: Optional[Dict[str, Any]],
                  schema: Optional[JsonObjectSchema],
                  name: str,
                  kind: str,
                  validator: Callable[[Dict[str, Any],
                                       JsonObjectSchema], Any]):
    if obj is None:
        return
    assert_instance(obj, dict, name=name)
    if schema is not None:
        assert_instance(schema, JsonObjectSchema,
                        name=f'{name}_schema')
        try:
            validator(obj, schema)
        except jsonschema.ValidationError as e:
            raise DataStoreError(f'Invalid {kind}'
                                 f' detected: {e.message}') from e



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
from typing import Union, Iterator, Set, Dict

import xarray as xr

from xcube.core.mldataset import MultiLevelDataset
from ..datasets.context import DatasetsContext
from ...util.assertions import assert_instance

_LEVELS_EXT = '.levels'
_ZARR_EXT = '.zarr'


class DatasetsMapping(collections.abc.Mapping):
    """Represents the given *datasets_ctx* as a mapping from
    dataset identifier to dataset, it can
    be passed to class:ObjectStorage.

    This is the applied Adapter design pattern to make
    class:DatasetsContext compatible with the mapping argument for
    class:ObjectStorage.

    The original identifiers will be renamed in case their suffixes
    do not match the desired bucket type, so the bucket contents are
    more user-friendly. If *is_multi_level* is True, the new names
    will always have ".levels" suffix, otherwise the ".zarr" suffix.

    :param datasets_ctx: The datasets' context
    :param is_multi_level: Whether this is a multi-level datasets'
        object storage
    """

    def __init__(self,
                 datasets_ctx: DatasetsContext,
                 is_multi_level: bool = False):
        assert_instance(datasets_ctx, DatasetsContext, name="datasets_ctx")
        assert_instance(is_multi_level, bool, name="is_multi_level")
        self._datasets_ctx = datasets_ctx
        self._is_multi_level = is_multi_level

    @property
    def _s3_names(self):
        return self._get_s3_names(self._datasets_ctx, self._is_multi_level)

    @staticmethod
    def _get_s3_names(datasets_ctx: DatasetsContext,
                      is_multi_level: bool) -> Dict[str, str]:
        """Generate user-friendly S3 names for dataset identifiers.
        If *is_multi_level* is True, S3 names will be forced to have
        ".levels" suffix, otherwise the ".zarr" suffix.
        """
        ds_ids = [c["Identifier"]
                  for c in datasets_ctx.get_dataset_configs()]
        all_ids = set(ds_ids)

        s3_names = {}
        for ds_id in ds_ids:
            s3_base, s3_ext = _split_base_ext(ds_id)
            if is_multi_level:
                s3_name = _replace_ext(s3_base, s3_ext,
                                       _ZARR_EXT, _LEVELS_EXT, all_ids)
            else:
                s3_name = _replace_ext(s3_base, s3_ext,
                                       _LEVELS_EXT, _ZARR_EXT, all_ids)
            s3_names[s3_name] = ds_id
        return s3_names

    def __len__(self) -> int:
        return len(self._s3_names)

    def __iter__(self) -> Iterator[str]:
        return iter(self._s3_names)

    def __contains__(self, s3_name: str) -> bool:
        """Check if *dataset_id* is a valid dataset.
        Overridden to avoid a call to __getitem__(),
        which will open the dataset (or raise ApiError!),
        but we want this to happen for direct __getitem__()
        calls only."""
        return s3_name in self._s3_names

    def __getitem__(self, s3_name: str) \
            -> Union[xr.Dataset, MultiLevelDataset]:
        """Get or open the dataset given by *dataset_id*."""
        dataset_id = self._s3_names[s3_name]
        # Will raise ApiError
        if self._is_multi_level:
            return self._datasets_ctx.get_ml_dataset(dataset_id)
        else:
            return self._datasets_ctx.get_dataset(dataset_id)


def _split_base_ext(identifier: str):
    base_ext = identifier.rsplit('.', maxsplit=1)
    if len(base_ext) == 2:
        return base_ext[0], '.' + base_ext[1]
    else:
        return identifier, ''


def _replace_ext(base_name: str, old_ext: str, trigger_ext: str, new_ext: str,
                 all_ids: Set[str]) -> str:
    if old_ext == new_ext \
            or (old_ext == trigger_ext
                and (base_name + new_ext) not in all_ids):
        return base_name + new_ext
    return base_name + old_ext + new_ext



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

UNDEFINED_STR = "UNDEFINED"


class _Undefined:
    """
    Represents the UNDEFINED value.
    """
    _hash_code = hash(UNDEFINED_STR) + 1

    def __str__(self):
        return UNDEFINED_STR

    def __repr__(self):
        return UNDEFINED_STR

    def __eq__(self, other):
        return self is other or isinstance(other, _Undefined)

    def __hash__(self) -> int:
        return _Undefined._hash_code


#: Singleton value used to indicate an undefined state.
UNDEFINED = _Undefined()



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Sequence

import click

from xcube.constants import (EXTENSION_POINT_DATASET_IOS,
                             LOG)
from xcube.constants import EXTENSION_POINT_INPUT_PROCESSORS
from xcube.constants import RESAMPLING_METHOD_NAMES
from xcube.core.gen.defaults import DEFAULT_OUTPUT_PATH, \
    DEFAULT_OUTPUT_RESAMPLING
from xcube.util.extension import Extension

resampling_methods = sorted(RESAMPLING_METHOD_NAMES)


# noinspection PyShadowingBuiltins
@click.command(name='gen', context_settings={"ignore_unknown_options": True})
@click.argument('input', nargs=-1)
@click.option('--proc', '-P', metavar='INPUT-PROCESSOR', default=None,
              help=f'Input processor name. '
                   f'The available input processor names and additional information about input processors '
                   'can be accessed by calling xcube gen --info . Defaults to "default", an input processor '
                   'that can deal with simple datasets whose variables have dimensions ("lat", "lon") and '
                   'conform with the CF conventions.')
@click.option('--config', '-c', metavar='CONFIG', multiple=True,
              help='xcube dataset configuration file in YAML format. More than one config input file is allowed.'
                   'When passing several config files, they are merged considering the order passed via command line.')
@click.option('--output', '-o', metavar='OUTPUT', default=DEFAULT_OUTPUT_PATH,
              help=f'Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}')
@click.option('--format', '-f', metavar='FORMAT',
              help=f'Output format. '
                   'Information about output formats can be accessed by calling '
                   'xcube gen --info. If omitted, the format will be guessed from the given output path.')
@click.option('--size', '-S', metavar='SIZE',
              help='Output size in pixels using format "<width>,<height>".')
@click.option('--region', '-R', metavar='REGION',
              help='Output region using format "<lon-min>,<lat-min>,<lon-max>,<lat-max>"')
@click.option('--variables', '--vars', metavar='VARIABLES',
              help='Variables to be included in output. '
                   'Comma-separated list of names which may contain wildcard characters "*" and "?".')
@click.option('--resampling', type=click.Choice(resampling_methods),
              default=DEFAULT_OUTPUT_RESAMPLING,
              help='Fallback spatial resampling algorithm to be used for all variables. '
                   f'Defaults to {DEFAULT_OUTPUT_RESAMPLING!r}. '
                   f'The choices for the resampling algorithm are: {resampling_methods}')
@click.option('--append', '-a', is_flag=True,
              help='Deprecated. The command will now always create, insert, replace, or append input slices.')
@click.option('--prof', is_flag=True,
              help='Collect profiling information and dump results after processing.')
@click.option('--no_sort', is_flag=True,
              help='The input file list will not be sorted before creating the xcube dataset. '
                   'If --no_sort parameter is passed, the order of the input list will be kept. '
                   'This parameter should be used for better performance, '
                   'provided that the input file list is in correct order (continuous time).')
@click.option('--info', '-I', is_flag=True,
              help='Displays additional information about format options or about input processors.')
@click.option('--dry_run', is_flag=True,
              help='Just read and process inputs, but don\'t produce any outputs.')
def gen(input: Sequence[str],
        proc: str,
        config: Sequence[str],
        output: str,
        format: str,
        size: str,
        region: str,
        variables: str,
        resampling: str,
        append: bool,
        prof: bool,
        dry_run: bool,
        info: bool,
        no_sort: bool):
    """
    Generate xcube dataset.
    Data cubes may be created in one go or successively for all given inputs.
    Each input is expected to provide a single time slice which may be appended, inserted or which may replace an
    existing time slice in the output dataset.
    The input paths may be one or more input files or a pattern that may contain wildcards '?', '*', and '**'.
    The input paths can also be passed as lines of a text file. To do so, provide exactly one input file with
    ".txt" extension which contains the actual input paths to be used.
    """
    dry_run = dry_run
    info_mode = info

    if info_mode:
        print(_format_info())
        return 0

    from xcube.core.gen.config import get_config_dict
    from xcube.core.gen.gen import gen_cube

    config = get_config_dict(
        input_paths=input,
        input_processor_name=proc,
        config_files=config,
        output_path=output,
        output_writer_name=format,
        output_size=size,
        output_region=region,
        output_variables=variables,
        output_resampling=resampling,
        profile_mode=prof,
        append_mode=append,
        no_sort_mode=no_sort,
    )

    gen_cube(dry_run=dry_run,
             monitor=LOG.info,
             **config)

    return 0


def _format_info():
    from xcube.util.plugin import get_extension_registry

    iproc_extensions = get_extension_registry().find_extensions(EXTENSION_POINT_INPUT_PROCESSORS)
    dsio_extensions = get_extension_registry().find_extensions(EXTENSION_POINT_DATASET_IOS,
                                                               lambda e: 'w' in e.metadata.get('modes', set()))

    help_text = '\nInput processors to be used with option --proc:\n'
    help_text += _format_input_processors(iproc_extensions)
    help_text += '\nFor more input processors use existing "xcube-gen-..." plugins ' \
                 "from the xcube's GitHub organisation or write your own plugin.\n"
    help_text += '\n'
    help_text += '\nOutput formats to be used with option --format:\n'
    help_text += _format_dataset_ios(dsio_extensions)
    help_text += '\n'

    return help_text


def _format_input_processors(input_processors: List[Extension]):
    help_text = ''
    for input_processor in input_processors:
        name = input_processor.name
        description = input_processor.metadata.get('description', '')
        fill = ' ' * (34 - len(input_processor.name))
        help_text += f'  {name}{fill}{description}\n'
    return help_text


def _format_dataset_ios(dataset_ios: List[Extension]):
    help_text = ''
    for ds_io in dataset_ios:
        name = ds_io.name
        description = ds_io.metadata.get('description', '')
        ext = ds_io.metadata.get('ext', '?')
        fill1 = ' ' * (24 - len(name))
        fill2 = ' ' * (10 - len(ext))
        help_text += f'  {name}{fill1}(*.{ext}){fill2}{description}\n'
    return help_text



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from abc import ABC, abstractmethod

from typing import Sequence, Tuple, Iterator, Generic, TypeVar, Any
from typing import Type, List
from typing import Union

import geopandas
import xarray

from xcube.core.mldataset import MultiLevelDataset
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonStringSchema

# A data type name or a DataType
DataTypeLike = Union[str, None, type, "DataType"]


class DataType:
    """
    A well-known Python data type that also has a representation
    using alias names.

    For example, the aliases "dataset" and "xarray.Dataset" both refer
    to the Python data type ``xarray.Dataset```.
    """

    _READTHEDOCS = os.environ.get("READTHEDOCS") == "True"
    _REGISTERED_DATA_TYPES: List["DataType"] = []

    @classmethod
    def register_data_type(cls, data_type: "DataType"):
        assert_instance(data_type, DataType, name="data_type")
        cls._REGISTERED_DATA_TYPES.append(data_type)

    def __init__(self, dtype: Type, alias: Union[None, str, Sequence[str]] = None):
        """
        :param dtype: The Python data type.
        :param alias: An alias name or list of aliases.
        """
        assert_instance(dtype, type if not self._READTHEDOCS else object, name="dtype")
        if alias is not None:
            assert_instance(alias, (str, tuple, list), name="alias")
        self._dtype = dtype
        self._aliases = (
            [] if alias is None else [alias] if isinstance(alias, str) else list(alias)
        ) + [self._get_fully_qualified_type_name(dtype)]
        self._alias_set = set(self._aliases)  # for faster lookup

    def __str__(self) -> str:
        return self.alias

    def __repr__(self) -> str:
        return repr(self.alias)

    def __eq__(self, other: Any):
        return self is other or (
            isinstance(other, DataType)
            and self.dtype is other.dtype
            and self.aliases == other.aliases
        )

    def __hash__(self):
        return hash(self.dtype) + 16 * hash(self.aliases)

    @property
    def dtype(self) -> Type:
        """The Python data type."""
        return self._dtype

    @property
    def alias(self) -> str:
        """The primary alias name."""
        return self._aliases[0]

    @property
    def aliases(self) -> Tuple[str, ...]:
        """All alias names."""
        return tuple(self._aliases)

    @classmethod
    def normalize(cls, data_type: DataTypeLike) -> "DataType":
        """
        Normalize the given *data_type* value into
        an instance of :class:DataType.

        :param data_type: Another data type, maybe given as type alias name,
            as a type, or as a DataType instance.
        :return: a DataType instance
        """
        if isinstance(data_type, DataType):
            return data_type
        if data_type is None or data_type is type(None):
            return ANY_TYPE
        if isinstance(data_type, str):
            for dt in cls._REGISTERED_DATA_TYPES:
                if data_type in dt._alias_set:
                    return dt
            raise ValueError(f"unknown data type {data_type!r}")
        if isinstance(data_type, type):
            for dt in cls._REGISTERED_DATA_TYPES:
                if data_type is dt.dtype:
                    return dt
            return DataType(data_type)
        raise ValueError(f"cannot convert {data_type!r} into a data type")

    def is_sub_type_of(self, data_type: DataTypeLike) -> bool:
        """
        Tests whether this data type is a sub-type of or the same
        as another data type.

        :param data_type: The other data type, may be given as type
            alias name, as a type, or as a DataType instance.
        :return: Whether this data type satisfies another data type.
        """
        return issubclass(self.dtype, self._normalize_dtype(data_type))

    def is_super_type_of(self, data_type: DataTypeLike) -> bool:
        """
        Tests whether this data type is a super-type of or the same
        as another data type.

        :param data_type: The other data type, may be given as type
            alias name, as a type, or as a DataType instance.
        :return: Whether this data type satisfies another data type.
        """
        return issubclass(self._normalize_dtype(data_type), self.dtype)

    @classmethod
    def get_schema(cls) -> JsonStringSchema:
        """Get the JSON Schema."""
        return JsonStringSchema(min_length=1, factory=cls.normalize, serializer=str)

    @classmethod
    def _normalize_dtype(cls, data_type: DataTypeLike) -> Type:
        if isinstance(data_type, DataType):
            return data_type.dtype
        if isinstance(data_type, type):
            return data_type
        return cls.normalize(data_type).dtype

    @staticmethod
    def _get_fully_qualified_type_name(data_type: Type) -> str:
        return f"{data_type.__module__}.{data_type.__name__}"


T = TypeVar("T")


class DataIterator(Generic[T], Iterator[T], ABC):
    """An iterator for data items.
    This class is a marker type for data stores that can return data iterators
    from their ``open_data()`` method.
    Data stores that support data iterators are not required to yield
    instances of this class. They just have to return iterator objects
    that yield instances of the desired item type, e.g. ``xarray.Dataset``.
    """

    @abstractmethod
    def __next__(self) -> T:
        """Yield the next data item."""


ANY_TYPE = DataType(object, ["any", "*", "object"])

DATASET_TYPE = DataType(xarray.Dataset, ["dataset", "xarray.Dataset"])

DATASET_ITERATOR_TYPE = DataType(
    DataIterator,  # Unfortunately we cannot use DataIterator[xarray.Dataset]
    ["dsiter", "DataIterator[xarray.Dataset]"],
)

MULTI_LEVEL_DATASET_TYPE = DataType(
    MultiLevelDataset,
    ["mldataset", "xcube.MultiLevelDataset", "xcube.core.mldataset.MultiLevelDataset"],
)

MULTI_LEVEL_DATASET_ITERATOR_TYPE = DataType(
    DataIterator,  # Unfortunately we cannot use DataIterator[MultiLevelDataset]
    ["mldsiter", "DataIterator[xcube.MultiLevelDataset]"],
)

GEO_DATA_FRAME_TYPE = DataType(
    geopandas.GeoDataFrame, ["geodataframe", "geopandas.GeoDataFrame"]
)

GEO_DATA_FRAME_ITERATOR_TYPE = DataType(
    DataIterator,  # Unfortunately we cannot use DataIterator[GeoDataFrame]
    ["gdfiter", "DataIterator[geopandas.GeoDataFrame]"],
)


def register_default_data_types():
    for data_type in [
        ANY_TYPE,
        DATASET_TYPE,
        DATASET_ITERATOR_TYPE,
        MULTI_LEVEL_DATASET_TYPE,
        MULTI_LEVEL_DATASET_ITERATOR_TYPE,
        GEO_DATA_FRAME_TYPE,
        GEO_DATA_FRAME_ITERATOR_TYPE,
    ]:
        DataType.register_data_type(data_type)


register_default_data_types()



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
from typing import Any, Dict

from xcube.util.versions import get_xcube_versions
from xcube.version import version
from .context import MetaContext


def get_service_info(ctx: MetaContext) -> Dict[str, Any]:
    api_infos = []
    for other_api in ctx.apis:
        api_info = {
            "name": other_api.name,
            "version": other_api.version,
            "description": other_api.description
        }
        api_infos.append({k: v
                          for k, v in api_info.items()
                          if v is not None})

    # TODO (forman): once APIs are configurable, we should
    #   get the server name, description from configuration too
    return dict(
        name="xcube Server",
        description="xcube Server",
        version=version,
        versions=get_xcube_versions(),
        apis=api_infos,
        startTime=ctx.start_time,
        currentTime=ctx.current_time,
        updateTime=ctx.update_time,
        pid=os.getpid()
    )



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import Tuple, Sequence, Mapping, \
    Optional, Dict, Any, Union, Hashable

import dask.array
import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr

from xcube.core.geom import get_dataset_bounds
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.timecoord import get_end_time_from_attrs
from xcube.core.timecoord import get_start_time_from_attrs
from xcube.core.timecoord import get_time_range_from_data
from xcube.core.timecoord import remove_time_part_from_isoformat
from xcube.util.assertions import assert_given, assert_true
from xcube.util.assertions import assert_not_none
from xcube.util.ipython import register_json_formatter
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonDateSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from .datatype import ANY_TYPE
from .datatype import DATASET_TYPE
from .datatype import DataType
from .datatype import DataTypeLike
from .datatype import GEO_DATA_FRAME_TYPE
from .datatype import MULTI_LEVEL_DATASET_TYPE


# TODO: IMPORTANT: replace, reuse, or align with
#   xcube.core.schema.CubeSchema class
#   xcube.webapi.context.DatasetDescriptor type
#   responses of xcube.webapi.controllers.catalogue
# TODO: write tests
# TODO: validate params


def new_data_descriptor(data_id: str, data: Any, require: bool = False) \
        -> 'DataDescriptor':
    if isinstance(data, MultiLevelDataset):
        dataset_descriptor_kwargs = _get_common_dataset_descriptor_props(
            data_id,
            # Note The highest level should have same metadata
            # and maybe loads faster.
            # data.get_dataset(data.num_levels - 1)
            data.get_dataset(0)
        )
        return MultiLevelDatasetDescriptor(
            num_levels=data.num_levels,
            **dataset_descriptor_kwargs
        )

    if isinstance(data, xr.Dataset):
        dataset_descriptor_kwargs = _get_common_dataset_descriptor_props(
            data_id,
            data
        )
        return DatasetDescriptor(
            **dataset_descriptor_kwargs
        )

    if isinstance(data, gpd.GeoDataFrame):
        # TODO: implement me: data -> GeoDataFrameDescriptor
        return GeoDataFrameDescriptor(data_id=data_id)

    if not require:
        return DataDescriptor(data_id=data_id,
                              data_type=ANY_TYPE)

    raise NotImplementedError()


def _get_common_dataset_descriptor_props(
        data_id: str,
        dataset: Union[xr.Dataset, MultiLevelDataset]
) -> Dict[str, Any]:
    dims = {str(k): v for k, v in dataset.dims.items()}
    coords = _build_variable_descriptor_dict(dataset.coords)
    data_vars = _build_variable_descriptor_dict(dataset.data_vars)
    spatial_res = _determine_spatial_res(dataset)
    bbox = _determine_bbox(dataset)
    time_range = _determine_time_coverage(dataset)
    time_period = _determine_time_period(dataset)
    return dict(
        data_id=data_id,
        dims=dims,
        coords=coords,
        data_vars=data_vars,
        bbox=bbox,
        time_range=time_range,
        time_period=time_period,
        spatial_res=spatial_res,
        attrs=dataset.attrs
    )


class DataDescriptor(JsonObject):
    """
    A generic descriptor for any data.
    Also serves as a base class for more specific data descriptors.

    :param data_id: An identifier for the data
    :param data_type: A type specifier for the data
    :param crs: A coordinate reference system identifier,
        as an EPSG, PROJ or WKT string
    :param bbox: A bounding box of the data
    :param time_range: Start and end time delimiting this
        data's temporal extent
    :param time_period: The data's periodicity if it is
        evenly temporally resolved.
    :param open_params_schema: A JSON schema describing the
        parameters that may be used to open this data.
    """

    def __init__(self,
                 data_id: str,
                 data_type: DataTypeLike,
                 *,
                 crs: str = None,
                 bbox: Tuple[float, float, float, float] = None,
                 time_range: Tuple[Optional[str], Optional[str]] = None,
                 time_period: str = None,
                 open_params_schema: JsonObjectSchema = None,
                 **additional_properties):
        assert_given(data_id, 'data_id')
        if additional_properties:
            warnings.warn(f'Additional properties received;'
                          f' will be ignored: {additional_properties}')
        self.data_id = data_id
        self.data_type = DataType.normalize(data_type)
        self.crs = crs
        self.bbox = tuple(bbox) if bbox else None
        self.time_range = tuple(time_range) if time_range else None
        self.time_period = time_period
        self.open_params_schema = open_params_schema

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                data_id=JsonStringSchema(min_length=1),
                data_type=DataType.get_schema(),
                crs=JsonStringSchema(min_length=1),
                bbox=JsonArraySchema(items=[JsonNumberSchema(),
                                            JsonNumberSchema(),
                                            JsonNumberSchema(),
                                            JsonNumberSchema()]),
                time_range=JsonDateSchema.new_range(nullable=True),
                time_period=JsonStringSchema(min_length=1),
                open_params_schema=JsonObjectSchema(
                    additional_properties=True
                ),
            ),
            required=['data_id', 'data_type'],
            additional_properties=True,
            factory=cls)


class DatasetDescriptor(DataDescriptor):
    """
    A descriptor for a gridded, N-dimensional dataset represented
    by xarray.Dataset. Comprises a description of the data variables
    contained in the dataset.

    Regrading *time_range* and *time_period* parameters, please refer to
    https://github.com/dcs4cop/xcube/blob/master/docs/source/storeconv.md#date-time-and-duration-specifications

    :param data_id: An identifier for the data
    :param data_type: The data type of the data described
    :param crs: A coordinate reference system identifier,
        as an EPSG, PROJ or WKT string
    :param bbox: A bounding box of the data
    :param time_range: Start and end time delimiting this data's
        temporal extent
    :param time_period: The data's periodicity
        if it is evenly temporally resolved
    :param spatial_res: The spatial extent of a pixel in crs units
    :param dims: A mapping of the dataset's dimensions to their sizes
    :param coords: mapping of the dataset's data coordinate names
        to instances of :class:VariableDescriptor
    :param data_vars: A mapping of the dataset's variable names
        to instances of :class:VariableDescriptor
    :param attrs: A mapping containing arbitrary attributes of the dataset
    :param open_params_schema: A JSON schema describing the parameters
        that may be used to open this data
    """

    def __init__(self,
                 data_id: str,
                 *,
                 data_type: DataTypeLike = DATASET_TYPE,
                 crs: str = None,
                 bbox: Tuple[float, float, float, float] = None,
                 time_range: Tuple[Optional[str], Optional[str]] = None,
                 time_period: str = None,
                 spatial_res: float = None,
                 dims: Mapping[str, int] = None,
                 coords: Mapping[str, 'VariableDescriptor'] = None,
                 data_vars: Mapping[str, 'VariableDescriptor'] = None,
                 attrs: Mapping[Hashable, any] = None,
                 open_params_schema: JsonObjectSchema = None,
                 **additional_properties):
        super().__init__(data_id=data_id,
                         data_type=data_type,
                         crs=crs,
                         bbox=bbox,
                         time_range=time_range,
                         time_period=time_period,
                         open_params_schema=open_params_schema)
        assert_true(DATASET_TYPE.is_super_type_of(data_type)
                    or MULTI_LEVEL_DATASET_TYPE.is_super_type_of(data_type),
                    f'illegal data_type,'
                    f' must be compatible with {DATASET_TYPE!r}'
                    f' or {MULTI_LEVEL_DATASET_TYPE!r}')
        if additional_properties:
            warnings.warn(f'Additional properties received;'
                          f' will be ignored: {additional_properties}')
        self.dims = dict(dims) if dims else None
        self.spatial_res = spatial_res
        self.coords = coords if coords else None
        self.data_vars = data_vars if data_vars else None
        self.attrs = _attrs_to_json(attrs) if attrs else None

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            dims=JsonObjectSchema(
                additional_properties=JsonIntegerSchema(minimum=0)
            ),
            spatial_res=JsonNumberSchema(exclusive_minimum=0.0),
            coords=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            data_vars=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            attrs=JsonObjectSchema(
                additional_properties=True
            ),
        )
        schema.required = ['data_id', 'data_type']
        schema.additional_properties = False
        schema.factory = cls
        return schema


class VariableDescriptor(JsonObject):
    """
    A descriptor for dataset variable represented by
    xarray.DataArray instances.
    They are part of dataset descriptor for an gridded, N-dimensional
    dataset represented by
    xarray.Dataset.

    :param name: The variable name
    :param dtype: The data type of the variable.
    :param dims: A list of the names of the variable's dimensions.
    :param chunks: A list of the chunk sizes of the variable's dimensions
    :param attrs: A mapping containing arbitrary attributes of the variable
    """

    def __init__(self,
                 name: str,
                 dtype: str,
                 dims: Sequence[str],
                 *,
                 chunks: Sequence[int] = None,
                 attrs: Mapping[Hashable, any] = None,
                 **additional_properties):
        assert_given(name, 'name')
        assert_given(dtype, 'dtype')
        assert_not_none(dims, 'dims')
        if additional_properties:
            warnings.warn(f'Additional properties received;'
                          f' will be ignored: {additional_properties}')
        self.name = name
        self.dtype = dtype
        self.dims = tuple(dims)
        self.chunks = tuple(chunks) if chunks else None
        self.attrs = _attrs_to_json(attrs) if attrs else None

    @property
    def ndim(self) -> int:
        """Number of dimensions."""
        return len(self.dims)

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                name=JsonStringSchema(min_length=1),
                dtype=JsonStringSchema(min_length=1),
                dims=JsonArraySchema(items=JsonStringSchema(min_length=1)),
                chunks=JsonArraySchema(items=JsonIntegerSchema(minimum=0)),
                attrs=JsonObjectSchema(additional_properties=True),
            ),
            required=['name', 'dtype', 'dims'],
            additional_properties=False,
            factory=cls)


class MultiLevelDatasetDescriptor(DatasetDescriptor):
    """
    A descriptor for a gridded, N-dimensional, multi-level,
    multi-resolution dataset represented by
    xcube.core.mldataset.MultiLevelDataset.

    :param data_id: An identifier of the multi-level dataset
    :param num_levels: The number of levels of this multi-level dataset
    :param data_type: A type specifier for the multi-level dataset
    """

    def __init__(self,
                 data_id: str,
                 num_levels: int,
                 *,
                 data_type: DataTypeLike = MULTI_LEVEL_DATASET_TYPE,
                 **kwargs):
        assert_given(data_id, 'data_id')
        assert_given(num_levels, 'num_levels')
        super().__init__(data_id=data_id, data_type=data_type, **kwargs)
        assert_true(MULTI_LEVEL_DATASET_TYPE.is_super_type_of(data_type),
                    f'illegal data_type,'
                    f' must be compatible with {MULTI_LEVEL_DATASET_TYPE!r}')
        self.num_levels = num_levels

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            num_levels=JsonIntegerSchema(minimum=1),
        )
        schema.required.append('num_levels')
        schema.additional_properties = False
        schema.factory = cls
        return schema


class GeoDataFrameDescriptor(DataDescriptor):
    """
    A descriptor for a geo-vector dataset represented by a
    geopandas.GeoDataFrame instance.

    :param data_id: An identifier of the geopandas.GeoDataFrame
    :param feature_schema: A schema describing the properties
        of the vector data
    :param kwargs: Parameters passed to super :class:DataDescriptor
    """

    def __init__(self,
                 data_id: str,
                 *,
                 data_type: DataTypeLike = GEO_DATA_FRAME_TYPE,
                 feature_schema: JsonObjectSchema = None,
                 **kwargs):
        super().__init__(data_id=data_id,
                         data_type=data_type,
                         **kwargs)
        assert_true(GEO_DATA_FRAME_TYPE.is_super_type_of(data_type),
                    f'illegal data_type,'
                    f' must be compatible with {GEO_DATA_FRAME_TYPE!r}')
        self.feature_schema = feature_schema

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            feature_schema=JsonObjectSchema(additional_properties=True),
        )
        schema.required = ['data_id']
        schema.additional_properties = False
        schema.factory = cls
        return schema


register_json_formatter(DataDescriptor)
register_json_formatter(DatasetDescriptor)
register_json_formatter(VariableDescriptor)
register_json_formatter(MultiLevelDatasetDescriptor)
register_json_formatter(GeoDataFrameDescriptor)


#############################################################################
# Implementation helpers


def _build_variable_descriptor_dict(variables) \
        -> Mapping[str, 'VariableDescriptor']:
    return {
        str(var_name):
            VariableDescriptor(
                name=str(var_name),
                dtype=str(var.dtype),
                dims=var.dims,
                chunks=tuple([max(chunk) for chunk in tuple(var.chunks)])
                if var.chunks else None,
                attrs=var.attrs
            )
        for var_name, var in variables.items()}


def _determine_bbox(data: xr.Dataset) \
        -> Optional[Tuple[float, float, float, float]]:
    try:
        return get_dataset_bounds(data)
    except ValueError:
        if 'geospatial_lon_min' in data.attrs and \
                'geospatial_lat_min' in data.attrs and \
                'geospatial_lon_max' in data.attrs and \
                'geospatial_lat_max' in data.attrs:
            return (data.geospatial_lon_min, data.geospatial_lat_min,
                    data.geospatial_lon_max, data.geospatial_lat_max)


def _determine_spatial_res(data: xr.Dataset):
    # TODO get rid of these hard-coded coord names as soon as
    #   new resampling is available
    lat_dimensions = ['lat', 'latitude', 'y']
    for lat_dimension in lat_dimensions:
        if lat_dimension in data:
            lat_diff = data[lat_dimension].diff(
                dim=data[lat_dimension].dims[0]
            ).values
            lat_res = lat_diff[0]
            lat_regular = np.allclose(lat_res, lat_diff, 1e-8)
            if lat_regular:
                return float(abs(lat_res))


def _determine_time_coverage(data: xr.Dataset):
    start_time, end_time = get_time_range_from_data(data)
    if start_time is not None:
        try:
            start_time = remove_time_part_from_isoformat(
                pd.to_datetime(start_time).isoformat()
            )
        except TypeError:
            start_time = None
    if start_time is None:
        start_time = get_start_time_from_attrs(data)
    if end_time is not None:
        try:
            end_time = remove_time_part_from_isoformat(
                pd.to_datetime(end_time).isoformat()
            )
        except TypeError:
            end_time = None
    if end_time is None:
        end_time = get_end_time_from_attrs(data)
    return start_time, end_time


def _determine_time_period(data: xr.Dataset):
    if 'time' in data and len(data['time'].values) > 1:
        time_diff = data['time'].diff(
            dim=data['time'].dims[0]
        ).values.astype(np.float64)
        time_res = time_diff[0]
        time_regular = np.allclose(time_res, time_diff, 1e-8)
        if time_regular:
            time_period = pd.to_timedelta(time_res).isoformat()
            # remove leading P
            time_period = time_period[1:]
            # removing sub-day precision
            return time_period.split('T')[0]


def _attrs_to_json(attrs: Mapping[Hashable, Any]) \
        -> Optional[Dict[str, Any]]:
    new_attrs: Dict[str, Any] = {}
    for k, v in attrs.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()
        elif isinstance(v, dask.array.Array):
            v = np.array(v).tolist()
        if isinstance(v, float) and np.isnan(v):
            v = None
        new_attrs[str(k)] = v
    return new_attrs



from abc import abstractmethod, ABC
from typing import Iterator

from xcube.util.jsonschema import JsonObjectSchema
from .assertions import assert_valid_params
from .datatype import DataTypeLike
from .descriptor import DataDescriptor


class DataSearcher(ABC):
    """
    Allow searching data in a data store.
    """

    @classmethod
    @abstractmethod
    def get_search_params_schema(cls,
                                 data_type: DataTypeLike = None) \
            -> JsonObjectSchema:
        """
        Get the schema for the parameters that can be passed
        as *search_params* to :meth:search_data().
        Parameters are named and described by the properties of the
        returned JSON object schema.

        :param data_type: If given, the search parameters
            will be tailored to search for data for the given *data_type*.
        :return: A JSON object schema whose properties describe
            this store's search parameters.
        """

    @abstractmethod
    def search_data(self,
                    data_type: DataTypeLike = None,
                    **search_params) -> Iterator[DataDescriptor]:
        """
        Search this store for data resources.
        If *data_type* is given, the search is restricted
        to data resources of that type.

        Returns an iterator over the search results which are
        returned as :class:DataDescriptor objects.

        If a store implementation supports only a single data type,
        it should verify that *data_type*
        is either None or compatible with the supported data type
        specifier.

        :param data_type: An optional data type
            that is known to be supported by this data store.
        :param search_params: The search parameters.
        :return: An iterator of data descriptors for the found
            data resources.
        :raise DataStoreError: If an error occurs.
        """


# noinspection PyUnresolvedReferences
class DefaultSearchMixin(DataSearcher):
    """
    A mixin for data store implementations that implements default
    search behaviour.

    It is expected that such data stores have no dedicated search
    parameters.
    """

    # noinspection PyUnusedLocal
    @classmethod
    def get_search_params_schema(cls, data_type: DataTypeLike = None) \
            -> JsonObjectSchema:
        """
        Get search parameters JSON object schema.

        The default implementation returns a schema that does
        not allow for any search parameters.

        :param data_type:
        :return: a JSON object schema for the search parameters
        """
        return JsonObjectSchema(additional_properties=False)

    def search_data(self,
                    data_type: DataTypeLike = None,
                    **search_params) \
            -> Iterator[DataDescriptor]:
        """
        Search the data store.

        The default implementation returns all data resources that
        may be filtered using the optional *data_type*.

        :param data_type: Type specifier to filter returned
            data resources.
        :param search_params: Not supported (yet)
        :return: an iterator of :class:DataDescriptor instances
        """
        search_params_schema = self.get_search_params_schema(
            data_type=data_type
        )
        assert_valid_params(search_params,
                            name='search_params',
                            schema=search_params_schema)
        for data_id in self.get_data_ids(data_type=data_type):
            data_descriptor = self.describe_data(
                data_id,
                data_type=data_type
            )
            yield data_descriptor



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click

DEFAULT_OUTPUT_PATH = 'out.csv'
DEFAULT_NUM_POINTS_MAX = 100


@click.command(name="genpts", hidden=True)
@click.argument('cube')
@click.option('--output', '-o', 'output_path',
              help=f'Output path. Defaults to "{DEFAULT_OUTPUT_PATH}".',
              default=DEFAULT_OUTPUT_PATH)
@click.option('--number', '-n', 'num_points_max',
              help=f'Number of point_data. Defaults to "{DEFAULT_NUM_POINTS_MAX}".',
              type=int,
              default=DEFAULT_NUM_POINTS_MAX)
def genpts(cube: str, output_path: str, num_points_max: int):
    """
    Generate synthetic data points from CUBE.

    Generates synthetic data points for a given data CUBE and write points to a CSV and GeoJSON file.
    The primary use of the tool is to point datasets for machine learning tasks
    and to create test points for the "xcube extract" command.
    """
    import numpy as np
    import pandas as pd
    import xarray as xr

    def _all_data_vars_valid(dataset: xr.DataArray):
        for _, var in dataset.data_vars.items():
            if not np.isfinite(var):
                return False
        return True

    cube = xr.open_zarr(cube)

    num_points = 0
    point_data = dict(time=[], lat=[], lon=[])
    for var_name in cube.data_vars:
        point_data[var_name] = []

    spatial_res = ((cube.lon[-1] - cube.lon[0]) / (cube.lon.size - 1)).values
    temporal_res = ((cube.time[-1] - cube.time[0]) // (cube.time.size - 1)).values

    while num_points < num_points_max:
        it = np.random.randint(0, cube.time.size)
        iy = np.random.randint(0, cube.lat.size)
        ix = np.random.randint(0, cube.lon.size)

        point = cube.isel(time=it, lat=iy, lon=ix)

        if _all_data_vars_valid(point):
            point_data['time'].append(point.time.values + 0.5 * temporal_res * np.random.logistic(scale=0.2))
            point_data['lat'].append(point.lat.values + 0.5 * spatial_res * np.random.logistic(scale=0.2))
            point_data['lon'].append(point.lon.values + 0.5 * spatial_res * np.random.logistic(scale=0.2))
            for var_name, var in point.data_vars.items():
                value = var.values
                if np.issubdtype(value.dtype, np.floating):
                    value += np.random.logistic(scale=0.05)
                point_data[var_name].append(value)
            num_points += 1

    if output_path.endswith('.csv'):
        pd.DataFrame(point_data).to_csv(output_path)
    elif output_path.endswith('.geojson'):
        import json
        with open(output_path, 'w') as fp:
            json.dump(_to_geojson_dict(point_data, num_points), fp, indent=2)


def _to_geojson_dict(point_data, num_points):
    import numpy as np

    features = []
    for i in range(num_points):
        x = float(point_data['lon'][i])
        y = float(point_data['lat'][i])
        properties = {}
        for k, v in point_data.items():
            if k in {'lon', 'lat'}:
                continue
            value = v[i]
            if np.issubdtype(value.dtype, np.floating):
                value = float(value)
            elif np.issubdtype(value.dtype, np.datetime64):
                value = str(value)
            else:
                value = int(value)
            properties[k] = value
        features.append({
            "type": "Feature",
            "id": i,
            "geometry": {
                "type": "Point",
                "coordinates": (x, y),
            },
            "properties": properties,
        })
    return {
        "type": "FeatureCollection",
        "features": features
    }


if __name__ == '__main__':
    genpts.main()



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import xarray as xr


def dump_dataset(dataset: xr.Dataset,
                 var_names=None,
                 show_var_encoding=False) -> str:
    """
    Dumps a dataset or variables into a text string.

    :param dataset: input dataset
    :param var_names: names of variables to be dumped
    :param show_var_encoding: also dump variable encodings?
    :return: the dataset dump
    """
    lines = []
    if not var_names:
        lines.append(str(dataset))
        if show_var_encoding:
            for var_name, var in dataset.coords.items():
                if var.encoding:
                    lines.append(dump_var_encoding(var, header=f"Encoding for coordinate variable {var_name!r}:"))
            for var_name, var in dataset.data_vars.items():
                if var.encoding:
                    lines.append(dump_var_encoding(var, header=f"Encoding for data variable {var_name!r}:"))
    else:
        for var_name in var_names:
            var = dataset[var_name]
            lines.append(str(var))
            if show_var_encoding and var.encoding:
                lines.append(dump_var_encoding(var))
    return "\n".join(lines)


def dump_var_encoding(var: xr.DataArray, header="Encoding:", indent=4) -> str:
    """
    Dump the encoding information of a variable into a text string.

    :param var: Dataset variable.
    :param header: Title/header string.
    :param indent: Indention in spaces.
    :return: the variable dump
    """
    lines = [header]
    max_len = 0
    for k in var.encoding:
        max_len = max(max_len, len(k))
    indent_spaces = indent * " "
    for k, v in var.encoding.items():
        tab_spaces = (2 + max_len - len(k)) * " "
        lines.append(f"{indent_spaces}{k}:{tab_spaces}{v!r}")
    return "\n".join(lines)



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Mapping, Sequence, Hashable, Dict

import numpy as np
import pandas as pd
import xarray as xr
from deprecated import deprecated

from xcube.core.schema import get_dataset_xy_var_names
from xcube.util.timeindex import ensure_time_label_compatible

# Exported for backward compatibility only
# noinspection PyUnresolvedReferences
from ._tile2 import (compute_tiles,
                     compute_rgba_tile,
                     get_var_valid_range,
                     get_var_cmap_params,
                     DEFAULT_VALUE_RANGE,
                     DEFAULT_FORMAT,
                     DEFAULT_CRS_NAME,
                     DEFAULT_TILE_SIZE,
                     DEFAULT_CMAP_NAME,
                     DEFAULT_TILE_ENLARGEMENT,
                     TileNotFoundException,
                     TileRequestException)


@deprecated(reason="Function is obsolete.",
            version="0.13.0")
def parse_non_spatial_labels(
        raw_labels: Mapping[str, str],
        dims: Sequence[Hashable],
        coords: Mapping[Hashable, xr.DataArray],
        allow_slices: bool = False,
        exception_type: type = ValueError,
        var: xr.DataArray = None
) -> Mapping[str, Any]:
    xy_var_names = get_dataset_xy_var_names(coords, must_exist=False)
    if xy_var_names is None:
        raise exception_type(f'missing spatial coordinates')
    xy_dims = set(coords[xy_var_name].dims[0] for xy_var_name in xy_var_names)

    # noinspection PyShadowingNames
    def to_datetime(datetime_str: str, dim_var: xr.DataArray):
        if datetime_str == 'current':
            return dim_var[-1]
        else:
            return pd.to_datetime(datetime_str)

    parsed_labels = {}
    for dim in dims:
        if dim in xy_dims:
            continue
        dim_var = coords[dim]
        label_str = raw_labels.get(str(dim))
        try:
            if label_str is None:
                label = dim_var[0].values
            elif label_str == 'current':
                label = dim_var[-1].values
            else:
                if '/' in label_str:
                    label_strs = tuple(label_str.split('/', maxsplit=1))
                else:
                    label_strs = (label_str,)
                if np.issubdtype(dim_var.dtype, np.floating):
                    labels = tuple(map(float, label_strs))
                elif np.issubdtype(dim_var.dtype, np.integer):
                    labels = tuple(map(int, label_strs))
                elif np.issubdtype(dim_var.dtype, np.datetime64):
                    labels = tuple(to_datetime(label, dim_var)
                                   for label in label_strs)
                else:
                    raise exception_type(f'unable to parse value'
                                         f' {label_str!r}'
                                         f' into a {dim_var.dtype!r}')
                if len(labels) == 1:
                    label = labels[0]
                else:
                    if allow_slices:
                        # noinspection PyTypeChecker
                        label = slice(labels[0], labels[1])
                    elif np.issubdtype(dim_var.dtype, np.integer):
                        label = labels[0] + (labels[1] - labels[0]) // 2
                    else:
                        label = labels[0] + (labels[1] - labels[0]) / 2
            parsed_labels[str(dim)] = label
        except ValueError as e:
            raise exception_type(f'{label_str!r} is not a valid'
                                 f' value for dimension {dim!r}') from e

    if var is not None:
        return ensure_time_label_compatible(var, parsed_labels)
    else:
        return parsed_labels




# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import fractions
import math
from typing import Tuple, List, Optional, Union

import click

from xcube.constants import EARTH_EQUATORIAL_PERIMETER

_DEFAULT_MIN_LEVEL = 0
_DEFAULT_MAX_TILE = 2500
_DEFAULT_RES_DELTA = "2.5%"
_DEFAULT_SORT_BY = "R_D"
_DEFAULT_NUM_RESULTS = 10
_DEFAULT_LAT_COVERAGE = fractions.Fraction(180, 1)

_SORT_BY_KEYS_0 = ['R_D', 'R_NOM', 'R_DEN', 'R', 'H', 'H0', 'L']
_SORT_BY_KEYS_P = ["+" + k for k in _SORT_BY_KEYS_0]
_SORT_BY_KEYS_M = ["-" + k for k in _SORT_BY_KEYS_0]
_SORT_BY_KEYS = _SORT_BY_KEYS_0 + _SORT_BY_KEYS_P + _SORT_BY_KEYS_M


def find_close_resolutions(target_res: float,
                           delta_res: float,
                           coverage: Union[int, fractions.Fraction],
                           max_height_0: int = _DEFAULT_MAX_TILE,
                           min_level: int = _DEFAULT_MIN_LEVEL,
                           int_inv_res: bool = False,
                           sort_by: str = _DEFAULT_SORT_BY) -> List[Tuple]:
    if target_res <= 0.0:
        raise ValueError('illegal target_res')
    if delta_res < 0.0 or delta_res >= target_res:
        raise ValueError('illegal delta_res')
    if min_level < 0.0:
        raise ValueError('illegal min_level')
    header = ("R_D (%)", "R_NOM", "R_DEN", "R (deg)", "R (m)", "H", "H0", "L")
    reverse_sort = False
    if sort_by.startswith("+") or sort_by.startswith("-"):
        reverse_sort = sort_by[0] == "-"
        sort_by = sort_by[1:]
    if sort_by == "R_D":
        def sort_key(item):
            return abs(item[0])
    elif sort_by == "R_NOM":
        def sort_key(item):
            return item[1]
    elif sort_by == "R_DEN":
        def sort_key(item):
            return item[2]
    elif sort_by == "R":
        def sort_key(item):
            return item[3]
    elif sort_by == "H":
        def sort_key(item):
            return item[5]
    elif sort_by == "H0":
        def sort_key(item):
            return item[6]
    elif sort_by == "L":
        def sort_key(item):
            return item[7]
    else:
        raise ValueError(f'illegal sort key: {sort_by}')
    # Compute h_min, h_max, the range of possible integer heights
    h_min = int(math.floor(coverage / (target_res + delta_res)))
    h_max = int(math.ceil(coverage / (target_res - delta_res)))
    # Collect resolutions all possible integer heights
    data = []
    for height in range(h_min, h_max + 1):
        res = coverage / fractions.Fraction(height, 1)
        # We may only want resolutions whose inverse is integer, e.g. 1/12 degree
        if not int_inv_res or res.numerator == 1:
            res_d = float(res)
            delta = res_d - target_res
            # Only if we are within delta_res
            if abs(delta) <= delta_res:
                # Only if res * height = coverage
                if res * height == coverage:
                    height_0, level = factor_out_two(height)
                    # Only if tile size falls below max and level exceeds min
                    if height_0 <= max_height_0 and level >= min_level:
                        delta_p = _round(100 * delta / target_res, 1000)
                        res_m = _round(degrees_to_meters(res_d), 100)
                        data.append((delta_p, res.numerator, res.denominator, res_d, res_m, height, height_0, level))
    data = sorted(data, key=sort_key, reverse=reverse_sort)
    # noinspection PyTypeChecker
    return [header] + data


def get_levels(height: int,
               coverage: Union[int, fractions.Fraction],
               level_min: int = None) -> List[Tuple]:
    res = coverage / fractions.Fraction(height)
    height_0, level = factor_out_two(height)
    data = []
    f = 1
    res_0 = res * (2 ** level)
    for i in range(0, max(level, level_min or level) + 1):
        height_i = height_0 * f
        res_i = res_0 / f
        res_i_d = float(res_i)
        res_i_m = _round(degrees_to_meters(res_i_d), 100)
        data.append((i, height_i, res_i, res_i_d, res_i_m))
        f *= 2
    header = ("L", "H", "R", "R (deg)", "R (m)")
    # noinspection PyTypeChecker
    return [header] + data


def get_adjusted_box(x1: float, y1: float, x2: float, y2: float, res: float) \
        -> Tuple[float, float, float, float]:
    # TODO: clamp values
    adj_x1 = res * math.floor(x1 / res)
    adj_y1 = res * math.floor(y1 / res)
    adj_x2 = res * math.ceil(x2 / res)
    adj_y2 = res * math.ceil(y2 / res)
    if adj_x2 - res >= x2:
        adj_x2 -= res
    if adj_y2 - res >= y2:
        adj_y2 -= res
    return adj_x1, adj_y1, adj_x2, adj_y2


def meters_to_degrees(res):
    return (360.0 * res) / EARTH_EQUATORIAL_PERIMETER


def degrees_to_meters(res):
    return (res / 360.0) * EARTH_EQUATORIAL_PERIMETER


def _round(x: float, n: int) -> float:
    return round(n * x) / n


def factor_out_two(x: int) -> Tuple[int, int]:
    if x < 0:
        raise ValueError("x must not be negative")
    if x == 0:
        return 0, 0
    e = 0
    while x % 2 == 0:
        x >>= 1
        e += 1
    return x, e


@click.command(name="res")
@click.argument('target_res', metavar='TARGET_RES')
@click.option('--delta_res', '-D', metavar='DELTA_RES', default=_DEFAULT_RES_DELTA,
              help=f'Maximum resolution delta. Defaults to {_DEFAULT_RES_DELTA}.')
@click.option('--coverage', '--cov', metavar="COVERAGE", default=str(_DEFAULT_LAT_COVERAGE),
              help=f'The vertical coverage in degrees. Defaults to {_DEFAULT_LAT_COVERAGE} degrees.')
@click.option('--tile_max', metavar='TILE_MAX', default=_DEFAULT_MAX_TILE, type=int,
              help=f'Maximum tile size. Defaults to {_DEFAULT_MAX_TILE}.')
@click.option('--level_min', '-l', metavar='LEVEL_MIN', default=_DEFAULT_MIN_LEVEL, type=int,
              help=f'Minimum resolution level. Defaults to {_DEFAULT_MIN_LEVEL}.')
@click.option('--int_inv_res', metavar='INT_INV_RES', is_flag=True,
              help=f'Find only resolutions whose inverse are integers.')
@click.option('--sort_by', metavar='SORT_BY',
              type=click.Choice(_SORT_BY_KEYS), default=_DEFAULT_SORT_BY,
              help='Sort output by column name.')
@click.option('--num_results', '-N', metavar='NUM_RESULTS',
              type=int, default=_DEFAULT_NUM_RESULTS,
              help=f'Maximum number of results to list. Defaults to {_DEFAULT_NUM_RESULTS}')
@click.option('--sep', metavar='SEP', default='\t',
              help='Column separator for the output. Defaults to TAB.')
def list_resolutions(target_res: str,
                     delta_res: str,
                     coverage: str,
                     tile_max: int,
                     level_min: int,
                     int_inv_res: bool,
                     sort_by: str,
                     num_results: int,
                     sep: str):
    """
    List resolutions close to a target resolution.

    Lists possible resolutions of a fixed Earth grid that are close to a given target
    resolution TARGET_RES within a maximum allowed deviation DELTA_RES.

    Both TARGET_RES and DELTA_RES can be suffixed by a "m" to indicate meter units.
    DELTA_RES can also be suffixed by a "%" to indicate deviation from TARGET_RES in percent.

    If LEVEL_MIN is provided and greater zero, only resolutions are listed whose
    HEIGHT is larger than TILE * 2 ^ LEVEL_MIN.
    """
    if target_res.endswith("m"):
        target_res = meters_to_degrees(float(target_res[0: -1]))
    else:
        target_res = float(target_res)

    if delta_res.endswith("m"):
        delta_res = meters_to_degrees(float(delta_res[0: -1]))
    elif delta_res.endswith("%"):
        delta_res = float(delta_res[0: -1]) * target_res / 100
    else:
        delta_res = float(delta_res)

    coverage = _fetch_coverage_from_option(coverage)

    sep = '\t' if sep.upper() == "TAB" else sep

    results = find_close_resolutions(target_res,
                                     delta_res,
                                     coverage,
                                     max_height_0=tile_max,
                                     min_level=level_min,
                                     int_inv_res=int_inv_res,
                                     sort_by=sort_by)

    click.echo()
    if len(results) <= num_results:
        for result in results:
            click.echo(sep.join(map(str, result)))
    else:
        for result in results[0: num_results]:
            click.echo(sep.join(map(str, result)))
        click.echo(f"{len(results) - num_results} more...")


@click.command(name="levels")
@click.option('--res', '-R', metavar="RES",
              help='Resolution in degrees. Can also be a rational number of form RES_NOM/RES_DEN.')
@click.option('--height', '-h', metavar="HEIGHT", type=int,
              help='Height in grid cells.')
@click.option('--coverage', '--cov', metavar="COVERAGE", default=str(_DEFAULT_LAT_COVERAGE),
              help=f'The vertical coverage in degrees. Defaults to {_DEFAULT_LAT_COVERAGE} degrees.')
@click.option('--level_min', '-l', metavar="LEVEL_MIN", type=int,
              help="List at least up to this level.")
@click.option('--sep', metavar='SEP', default='\t',
              help='Column separator for the output. Defaults to TAB.')
def list_levels(res: str, height: int, coverage: str, level_min: Optional[int], sep: str):
    """
    List levels for a resolution or a tile size.

    Lists the given number of LEVELS for given resolution RES or given height in grid cells HEIGHT.
    which can both be used to define a fixed Earth grid.
    """
    height, coverage = _fetch_height_and_coverage_from_options(res, height, coverage)

    sep = '\t' if sep.upper() == "TAB" else sep

    rows = get_levels(height, coverage, level_min)

    click.echo()
    for row in rows:
        click.echo(sep.join(map(str, row)))


@click.command(name="abox", context_settings={"ignore_unknown_options": True})
@click.argument('geom', metavar="GEOM")
@click.option('--res', '-R', metavar="RES",
              help='The parent grid\'s resolution in degrees. Can also be a rational number of form A/B.')
@click.option('--height', '-h', metavar="HEIGHT", type=int,
              help='The parent grid\'s height in grid cells.')
@click.option('--coverage', '--cov', metavar="COVERAGE", default=str(_DEFAULT_LAT_COVERAGE),
              help=f'The parent grid\'s coverage in degrees. Defaults to {_DEFAULT_LAT_COVERAGE} degrees.')
@click.option('--tile_factor', metavar="TILE_FACTOR", type=int,
              help='A tile factor to compute tile sizes from height at level zero: TILE = TILE_FACTOR * HEIGHT_0.'
                   'Usually TILE_FACTOR = 2^N. If not given, TILE = 1.')
def adjust_box(geom: str, res: Optional[str], height: Optional[int], coverage: str, tile_factor: Optional[int]):
    """
    Adjust a bounding box to a fixed Earth grid.

    Adjusts a bounding box given by GEOM  to a fixed Earth grid by the
    inverse resolution INV_RES in degrees^-1 units, which must be an integer number.

    GEOM is a bounding box given as x1,y1,x2,y2 in decimal degrees.
    (Geometry WKT and GeoJSON support may be added later.)
    """
    try:
        x1, y1, x2, y2 = [float(c) for c in geom.split(",")]
    except (ValueError, TypeError) as e:
        raise click.ClickException(f"Invalid GEOM: {geom}") from e

    height, coverage = _fetch_height_and_coverage_from_options(res, height, coverage)

    height_0, level = factor_out_two(height)

    res = coverage / fractions.Fraction(height, 1)
    res_0 = res * 2 ** level

    if tile_factor is not None:
        tile_size = tile_factor * height_0
    else:
        tile_size = 1
    graticule_dist = res_0 * tile_size

    # Adjust along tile boundaries
    adj_x1, adj_y1, adj_x2, adj_y2 = get_adjusted_box(x1, y1, x2, y2, float(graticule_dist))

    reg_width = round((adj_x2 - adj_x1) / float(res))
    reg_height = round((adj_y2 - adj_y1) / float(res))

    orig_coords = (f'(({x1} {y1},'
                   f' {x2} {y1},'
                   f' {x2} {y2},'
                   f' {x1} {y2},'
                   f' {x1} {y1}))')
    adj_coords = (f'(({adj_x1} {adj_y1},'
                  f' {adj_x2} {adj_y1},'
                  f' {adj_x2} {adj_y2},'
                  f' {adj_x1} {adj_y2},'
                  f' {adj_x1} {adj_y1}))')

    click.echo()
    click.echo(f'Orig. box coord. = {x1},{y1},{x2},{y2}')
    click.echo(f'Adj. box coord.  = {adj_x1},{adj_y1},{adj_x2},{adj_y2}')
    click.echo(f'Orig. box WKT    = POLYGON {orig_coords}')
    click.echo(f'Adj. box WKT     = POLYGON {adj_coords}')
    click.echo(f'Combined WKT     = MULTIPOLYGON ({orig_coords}, {adj_coords})')
    click.echo(f'Box grid size    = {reg_width} x {reg_height} cells')
    click.echo(f'Graticule dist.  = {graticule_dist} degrees')
    click.echo(f'Tile size        = {tile_size} cells')
    click.echo(f'Granularity      = {height_0} cells')
    click.echo(f'Level            = {level}')
    click.echo(f'Res. at level 0  = {res_0} degrees')
    click.echo(f'Resolution       = {res} degrees')
    click.echo(f'                 = {_round(degrees_to_meters(res), 100)} meters')


def _fetch_height_and_coverage_from_options(res_str: Optional[str],
                                            height: Optional[int],
                                            coverage_str: str) -> Tuple[int, fractions.Fraction]:
    coverage = _fetch_coverage_from_option(coverage_str)
    if res_str is not None:
        if height is not None:
            raise click.ClickException(f"Either RES or HEIGHT must be given, not both")
        try:
            res = fractions.Fraction(res_str)
        except ValueError as e:
            raise click.ClickException(f"Invalid RES: {res_str}") from e
        if res <= 0:
            raise click.ClickException(f"Invalid RES: {res_str}")
        height = coverage / res
        if height.denominator != 1:
            raise click.ClickException(f"Invalid RES: {res_str}, {coverage_str}/RES must be an integer number.")
        height = height.numerator
    elif height is None:
        raise click.ClickException(f"Either RES or HEIGHT must be given.")
    return height, coverage


def _fetch_coverage_from_option(coverage_str: str) -> fractions.Fraction:
    try:
        coverage = fractions.Fraction(coverage_str)
    except ValueError as e:
        raise click.ClickException(f"Invalid COVERAGE: {coverage_str}") from e
    if coverage <= 0:
        raise click.ClickException(f"Invalid COVERAGE: {coverage_str}")

    return coverage


@click.group()
def grid():
    """
    Find spatial xcube dataset resolutions and adjust bounding boxes.

    We find suitable resolutions with respect to a possibly regional fixed Earth grid and adjust regional spatial
    bounding boxes to that grid. We also try to select the resolutions such
    that they are taken from a certain level of a multi-resolution pyramid whose
    level resolutions increase by a factor of two.

    The graticule at a given resolution level L within the grid is given by

    \b
        RES(L) = COVERAGE * HEIGHT(L)
        HEIGHT(L) = HEIGHT_0 * 2 ^ L
        LON(L, I) = LON_MIN + I * HEIGHT_0 * RES(L)
        LAT(L, J) = LAT_MIN + J * HEIGHT_0 * RES(L)

    With

    \b
        RES:      Grid resolution in degrees.
        HEIGHT:   Number of vertical grid cells for given level
        HEIGHT_0: Number of vertical grid cells at lowest resolution level.

    Let WIDTH and HEIGHT be the number of horizontal and vertical grid cells
    of a global grid at a certain LEVEL with WIDTH * RES = 360 and HEIGHT * RES = 180, then
    we also force HEIGHT = TILE * 2 ^ LEVEL.
   """


grid.add_command(list_resolutions)
grid.add_command(list_levels)
grid.add_command(adjust_box)



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
import json
import math
import pathlib
import warnings
from functools import cached_property
from typing import Any, Optional, List, Union, Mapping, Dict, Sequence

import fsspec
import fsspec.core
import numpy as np
import xarray as xr
import zarr

# noinspection PyUnresolvedReferences
import xcube.core.zarrstore
from xcube.core.gridmapping import GridMapping
from xcube.core.subsampling import AggMethods, AggMethod
from xcube.util.assertions import assert_instance
from xcube.util.fspath import get_fs_path_class
from xcube.util.fspath import resolve_path
from xcube.util.types import ScalarOrPair
from xcube.util.types import normalize_scalar_or_pair
from .abc import MultiLevelDataset
from .base import BaseMultiLevelDataset
from .lazy import LazyMultiLevelDataset

LEVELS_FORMAT_VERSION = "1.0"


class FsMultiLevelDataset(LazyMultiLevelDataset):
    _MIN_CACHE_SIZE = 1024 * 1024  # 1 MiB

    def __init__(self,
                 path: str,
                 fs: Optional[fsspec.AbstractFileSystem] = None,
                 fs_root: Optional[str] = None,
                 fs_kwargs: Optional[Mapping[str, Any]] = None,
                 cache_size: Optional[int] = None,
                 consolidate: Optional[bool] = None,
                 **zarr_kwargs):
        if fs is None:
            fs, path = fsspec.core.url_to_fs(path, **(fs_kwargs or {}))
        assert_instance(fs, fsspec.AbstractFileSystem, name="fs")
        assert_instance(path, str, name="data_id")
        super().__init__(ds_id=path)
        self._path = path
        self._fs = fs
        self._fs_root = fs_root
        self._cache_size = cache_size
        self._consolidate = consolidate
        self._zarr_kwargs = zarr_kwargs
        self._path_class = get_fs_path_class(fs)

    @property
    def path(self) -> str:
        return self._path

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        return self._fs

    @property
    def cache_size(self) -> Optional[int]:
        return self._cache_size

    @cached_property
    def size_weights(self) -> np.ndarray:
        """Size weights are used to distribute the cache size
        over the levels.
        """
        return self.compute_size_weights(self.num_levels)

    @cached_property
    def tile_size(self) -> Optional[Sequence[int]]:
        spec = self._levels_spec
        return spec.get("tile_size")

    @cached_property
    def use_saved_levels(self) -> Optional[bool]:
        spec = self._levels_spec
        return spec.get("use_saved_levels")

    @cached_property
    def base_dataset_path(self) -> Optional[str]:
        spec = self._levels_spec
        return spec.get("base_dataset_path")

    @cached_property
    def agg_methods(self) -> Optional[Mapping[str, AggMethod]]:
        spec = self._levels_spec
        return spec.get("agg_methods")

    @cached_property
    def _levels_spec(self) -> Mapping[str, Any]:
        path = f"{self._path}/.zlevels"
        spec = {}
        if self.fs.exists(path):
            with self.fs.open(f"{self._path}/.zlevels") as fp:
                spec = json.load(fp)
            if not isinstance(spec, collections.abc.Mapping):
                raise TypeError('Unexpected .zlevels file.'
                                ' Must be a JSON object.')
        # TODO (forman): validate JSON object
        return spec

    def _get_dataset_lazily(self, index: int, parameters) \
            -> xr.Dataset:

        cache_size = self._cache_size

        fs = self._fs

        ds_path = self._get_path(self._path)
        link_path = ds_path / f'{index}.link'
        if fs.isfile(str(link_path)):
            # If file "{index}.link" exists, we have a link to
            # a level Zarr and open this instead,
            with fs.open(str(link_path), 'r') as fp:
                level_path = self._get_path(fp.read())
            if not level_path.is_absolute() \
                    and not self._is_path_relative_to_path(level_path,
                                                           ds_path):
                level_path = resolve_path(ds_path / level_path)
        else:
            # Nominal "{index}.zarr" must exist
            level_path = ds_path / f'{index}.zarr'

        level_zarr_store = fs.get_mapper(str(level_path))

        consolidated = self._consolidate if self._consolidate is not None \
            else ('.zmetadata' in level_zarr_store)

        if isinstance(cache_size, int) \
                and cache_size >= self._MIN_CACHE_SIZE:
            # compute cache size for level weighted by
            # size in pixels for each level
            cache_size = math.ceil(self.size_weights[index] * cache_size)
            if cache_size >= self._MIN_CACHE_SIZE:
                level_zarr_store = zarr.LRUStoreCache(level_zarr_store,
                                                      max_size=cache_size)

        try:
            level_dataset = xr.open_zarr(level_zarr_store,
                                         consolidated=consolidated,
                                         **self._zarr_kwargs)
        except ValueError as e:
            raise FsMultiLevelDatasetError(
                f'Failed to open'
                f' dataset {level_path!r}:'
                f' {e}'
            ) from e

        level_dataset.zarr_store.set(level_zarr_store)
        return level_dataset

    @staticmethod
    def _is_path_relative_to_path(level_path, ds_path):
        if hasattr(level_path, 'is_relative_to'):
            # Python >=3.9
            return level_path.is_relative_to(ds_path)
        try:
            # Python <3.9
            level_path.relative_to(ds_path)
            return True
        except ValueError:
            return False

    @classmethod
    def compute_size_weights(cls, num_levels: int) -> np.ndarray:
        weights = (2 ** np.arange(0, num_levels, dtype=np.float64)) ** 2
        return weights[::-1] / np.sum(weights)

    def _get_num_levels_lazily(self) -> int:
        spec = self._levels_spec
        num_levels = spec.get("num_levels")
        levels = self._get_levels()
        if num_levels is None:
            num_levels = len(levels)
        expected_levels = list(range(num_levels))
        for level in expected_levels:
            if level != levels[level]:
                raise FsMultiLevelDatasetError(
                    f'Inconsistent'
                    f' multi-level dataset {self.ds_id!r},'
                    f' expected levels {expected_levels!r}'
                    f' found {levels!r}'
                )
        return num_levels

    def _get_levels(self) -> List[int]:
        levels = []
        paths = [self._get_path(entry['name'])
                 for entry in self._fs.listdir(self._path, detail=True)]
        for path in paths:
            # No ext, i.e. dir_name = "<level>", is proposed by
            # https://github.com/zarr-developers/zarr-specs/issues/50.
            # xcube already selected dir_name = "<level>.zarr".
            basename = path.stem
            if path.stem and path.suffix in ('', '.zarr', '.link'):
                try:
                    level = int(basename)
                except ValueError:
                    continue
                levels.append(level)
        levels = sorted(levels)
        return levels

    def _get_path(self, *args) -> pathlib.PurePath:
        return self._path_class(*args)

    @classmethod
    def write_dataset(cls,
                      dataset: Union[xr.Dataset, MultiLevelDataset],
                      path: str,
                      fs: Optional[fsspec.AbstractFileSystem] = None,
                      fs_root: Optional[str] = None,
                      fs_kwargs: Optional[Mapping[str, Any]] = None,
                      replace: bool = False,
                      num_levels: Optional[int] = None,
                      consolidated: bool = True,
                      tile_size: Optional[ScalarOrPair[int]] = None,
                      use_saved_levels: bool = False,
                      base_dataset_path: Optional[str] = None,
                      agg_methods: Optional[AggMethods] = None,
                      **zarr_kwargs) -> str:

        assert_instance(dataset, (xr.Dataset, MultiLevelDataset),
                        name='dataset')

        if fs is None:
            fs, path = fsspec.core.url_to_fs(path, **(fs_kwargs or {}))
        if tile_size is not None:
            tile_size = normalize_scalar_or_pair(tile_size,
                                                 item_type=int,
                                                 name='tile_size')

        assert_instance(path, str, name='path')
        assert_instance(fs, fsspec.AbstractFileSystem, name='fs')

        if isinstance(dataset, MultiLevelDataset):
            ml_dataset = dataset
            if tile_size:
                warnings.warn(
                    'tile_size is ignored for multi-level datasets'
                )
            if agg_methods:
                warnings.warn(
                    'agg_methods is ignored for multi-level datasets'
                )
        else:
            base_dataset: xr.Dataset = dataset
            grid_mapping = None
            if tile_size is not None:
                grid_mapping = GridMapping.from_dataset(base_dataset)
                x_name, y_name = grid_mapping.xy_dim_names
                # noinspection PyTypeChecker
                base_dataset = base_dataset.chunk({x_name: tile_size[0],
                                                   y_name: tile_size[1]})
                # noinspection PyTypeChecker
                grid_mapping = grid_mapping.derive(tile_size=tile_size)
            ml_dataset = BaseMultiLevelDataset(base_dataset,
                                               grid_mapping=grid_mapping,
                                               num_levels=num_levels,
                                               agg_methods=agg_methods)

        if use_saved_levels:
            ml_dataset = BaseMultiLevelDataset(
                ml_dataset.get_dataset(0),
                grid_mapping=ml_dataset.grid_mapping,
                agg_methods=agg_methods
            )

        path_class = get_fs_path_class(fs)
        data_path = path_class(path)
        fs.mkdirs(str(data_path), exist_ok=replace)

        if num_levels is None or num_levels <= 0:
            num_levels_max = ml_dataset.num_levels
        else:
            num_levels_max = min(num_levels, ml_dataset.num_levels)

        with fs.open(str(data_path / '.zlevels'), mode='w') as fp:
            levels_data: Dict[str, Any] = dict(
                version=LEVELS_FORMAT_VERSION,
                num_levels=num_levels_max
            )
            if use_saved_levels is not None:
                levels_data.update(use_saved_levels=bool(use_saved_levels))
            if base_dataset_path:
                levels_data.update(base_dataset_path=base_dataset_path)
            if tile_size is not None:
                levels_data.update(tile_size=list(tile_size))
            if hasattr(ml_dataset, 'agg_methods'):
                levels_data.update(agg_methods=dict(ml_dataset.agg_methods))
            json.dump(levels_data, fp, indent=2)

        for index in range(num_levels_max):
            level_dataset = ml_dataset.get_dataset(index)
            if base_dataset_path and index == 0:
                assert_instance(fs_root, str, name="fs_root")

                # Write file "0.link" instead of copying
                # level zero dataset to "0.zarr".

                # Compute a relative base dataset path first
                base_dataset_path = path_class(fs_root, base_dataset_path)
                data_parent_path = data_path.parent
                try:
                    base_dataset_path = base_dataset_path.relative_to(
                        data_parent_path
                    )
                except ValueError as e:
                    raise FsMultiLevelDatasetError(
                        f'Invalid base_dataset_id: {base_dataset_path}'
                    ) from e
                base_dataset_path = '..' / base_dataset_path

                # Then write relative base dataset path into link file
                link_path = data_path / f'{index}.link'
                with fs.open(str(link_path), mode='w') as fp:
                    fp.write(base_dataset_path.as_posix())
            else:
                # Write level "{index}.zarr"
                level_path = data_path / f'{index}.zarr'
                level_zarr_store = fs.get_mapper(str(level_path), create=True)
                try:
                    level_dataset.to_zarr(
                        level_zarr_store,
                        mode='w' if replace else None,
                        consolidated=consolidated,
                        **zarr_kwargs
                    )
                except ValueError as e:
                    # TODO: remove already written data!
                    raise FsMultiLevelDatasetError(
                        f'Failed to write dataset {path}: {e}'
                    ) from e
                if use_saved_levels:
                    level_dataset = xr.open_zarr(level_zarr_store,
                                                 consolidated=consolidated)
                    level_dataset.zarr_store.set(level_zarr_store)
                    ml_dataset.set_dataset(index, level_dataset)

        return path


class FsMultiLevelDatasetError(ValueError):

    def __init__(self, message: str):
        super().__init__(message)



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

import click

from xcube.cli.common import (cli_option_scheduler,
                              cli_option_traceback,
                              handle_cli_exception,
                              new_cli_ctx_obj,
                              configure_logging,
                              configure_warnings)
from xcube.constants import (EXTENSION_POINT_CLI_COMMANDS,
                             LOG_LEVELS,
                             LOG_LEVEL_OFF_NAME)
from xcube.util.plugin import get_extension_registry
from xcube.version import version


# noinspection PyShadowingBuiltins,PyUnusedLocal
@click.group(name='xcube')
@click.version_option(version)
@cli_option_traceback
@cli_option_scheduler
@click.option('--loglevel', 'log_level',
              metavar='LOG_LEVEL',
              type=click.Choice(LOG_LEVELS),
              default=LOG_LEVEL_OFF_NAME,
              help=f'Log level.'
                   f' Must be one of {", ".join(LOG_LEVELS)}.'
                   f' Defaults to {LOG_LEVEL_OFF_NAME}.'
                   f' If the level is not {LOG_LEVEL_OFF_NAME},'
                   f' any log messages up to the given level will be'
                   f' written either to the console (stderr)'
                   f' or LOG_FILE, if provided.')
@click.option('--logfile', 'log_file',
              metavar='LOG_FILE',
              help=f'Log file path.'
                   f' If given, any log messages will redirected into'
                   f' LOG_FILE. Disables console output'
                   f' unless otherwise enabled, e.g.,'
                   f' using the --verbose flag.'
                   f' Effective only if LOG_LEVEL'
                   f' is not {LOG_LEVEL_OFF_NAME}.')
@click.option('--warnings', '-w',
              is_flag=True,
              help='Show any warnings emitted during operation'
                   ' (warnings are hidden by default).')
def cli(traceback=False,
        scheduler=None,
        log_level=None,
        log_file=None,
        warnings=None):
    """
    xcube Toolkit
    """
    configure_logging(log_file=log_file, log_level=log_level)
    configure_warnings(warnings)


_cli_commands_registered = False


def _register_cli_commands():
    global _cli_commands_registered
    if _cli_commands_registered:
        return
    _cli_commands_registered = True
    # Add registered CLI commands
    registry = get_extension_registry()
    for command in registry.find_components(EXTENSION_POINT_CLI_COMMANDS):
        cli.add_command(command)


_register_cli_commands()


def main(args=None):
    # noinspection PyBroadException
    ctx_obj = new_cli_ctx_obj()
    try:
        exit_code = cli.main(args=args, obj=ctx_obj, standalone_mode=False)
    except Exception as e:
        exit_code = handle_cli_exception(
            e, traceback_mode=ctx_obj.get("traceback", False)
        )
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import abc
from typing import Union, Callable, Optional, Any, \
    Awaitable, TypeVar

from tornado import concurrent

ReturnT = TypeVar("ReturnT")


class AsyncExecution(abc.ABC):
    """
    An interface that defines asynchronous function execution.
    """

    @abc.abstractmethod
    def call_later(self,
                   delay: Union[int, float],
                   callback: Callable,
                   *args,
                   **kwargs) -> object:
        """
        Executes the given callable *callback* after *delay* seconds.

        The method returns a handle that can be used to cancel the
        callback.

        :param delay: Delay in seconds.
        :param callback: Callback to be called.
        :param args: Positional arguments passed to *callback*.
        :param kwargs: Keyword arguments passed to *callback*.
        :return: A handle that provides the methods
            ``cancel()`` and ``cancelled()``.
        """

    @abc.abstractmethod
    def run_in_executor(
            self,
            executor: Optional[concurrent.futures.Executor],
            function: Callable[..., ReturnT],
            *args: Any,
            **kwargs: Any
    ) -> Awaitable[ReturnT]:
        """
        Concurrently runs a *function* in a ``concurrent.futures.Executor``.
        If *executor* is ``None``, the framework's default
        executor will be used.

        :param executor: An optional executor.
        :param function: The function to be run concurrently.
        :param args: Positional arguments passed to *function*.
        :param kwargs: Keyword arguments passed to *function*.
        :return: The awaitable return value of *function*.
        """



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import abstractmethod, ABC
from typing import Any, List, Optional

import xarray as xr

from xcube.constants import EXTENSION_POINT_DATA_OPENERS
from xcube.constants import EXTENSION_POINT_DATA_WRITERS
from xcube.util.assertions import assert_given
from xcube.util.extension import Extension
from xcube.util.extension import ExtensionPredicate
from xcube.util.extension import ExtensionRegistry
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.plugin import get_extension_registry
from .datatype import DataType
from .datatype import DataTypeLike
from .error import DataStoreError


#######################################################
# Data accessor instantiation and registry query
#######################################################

def new_data_opener(opener_id: str,
                    extension_registry: Optional[ExtensionRegistry] = None,
                    **opener_params) -> 'DataOpener':
    """
    Get an instance of the data opener identified by *opener_id*.

    The optional, extra opener parameters *opener_params* may
    be used by data store (``xcube.core.store.DataStore``)
    implementations so they can share their internal state with the opener.

    :param opener_id: The data opener identifier.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :param opener_params: Extra opener parameters.
    :return: A data opener instance.
    """
    assert_given(opener_id, 'opener_id')
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_OPENERS,
                                            opener_id):
        raise DataStoreError(f'A data opener named'
                             f' {opener_id!r} is not registered')
    return extension_registry.get_component(EXTENSION_POINT_DATA_OPENERS,
                                            opener_id)(**opener_params)


def new_data_writer(writer_id: str,
                    extension_registry: Optional[ExtensionRegistry] = None,
                    **writer_params) -> 'DataWriter':
    """
    Get an instance of the data writer identified by *writer_id*.

    The optional, extra writer parameters *writer_params* may be used by
    data store (``xcube.core.store.DataStore``) implementations so they
    can share their internal state with the writer.

    :param writer_id: The data writer identifier.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :param writer_params: Extra writer parameters.
    :return: A data writer instance.
    """
    assert_given(writer_id, 'writer_id')
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_WRITERS,
                                            writer_id):
        raise DataStoreError(f'A data writer named'
                             f' {writer_id!r} is not registered')
    return extension_registry.get_component(EXTENSION_POINT_DATA_WRITERS,
                                            writer_id)(**writer_params)


def find_data_opener_extensions(
        predicate: ExtensionPredicate = None,
        extension_registry: Optional[ExtensionRegistry] = None
) -> List[Extension]:
    """
    Get registered data opener extensions using the optional
    filter function *predicate*.

    :param predicate: An optional filter function.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :return: List of matching extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(
        EXTENSION_POINT_DATA_OPENERS,
        predicate=predicate
    )


def find_data_writer_extensions(
        predicate: ExtensionPredicate = None,
        extension_registry: Optional[ExtensionRegistry] = None
) -> List[Extension]:
    """
    Get registered data writer extensions using the optional filter
    function *predicate*.

    :param predicate: An optional filter function.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :return: List of matching extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(
        EXTENSION_POINT_DATA_WRITERS,
        predicate=predicate
    )


def get_data_accessor_predicate(
        data_type: DataTypeLike = None,
        format_id: str = None,
        storage_id: str = None
) -> ExtensionPredicate:
    """
    Get a predicate that checks if a data accessor extensions's name is
    compliant with *data_type*, *format_id*, *storage_id*.

    :param data_type: Optional data data type to be supported.
        May be given as type alias name, as a type,
        or as a DataType instance.
    :param format_id: Optional data format identifier to be supported.
    :param storage_id: Optional data storage identifier to be supported.
    :return: A filter function.
    :raise DataStoreError: If an error occurs.
    """
    if any((data_type, format_id, storage_id)):
        data_type = DataType.normalize(data_type) \
            if data_type is not None else None

        def _predicate(extension: Extension) -> bool:
            extension_parts = extension.name.split(':', maxsplit=4)
            if storage_id is not None:
                ext_storage_id = extension_parts[2]
                if ext_storage_id != '*' and ext_storage_id != storage_id:
                    return False
            if format_id is not None:
                ext_format_id = extension_parts[1]
                if ext_format_id != '*' and ext_format_id != format_id:
                    return False
            if data_type is not None:
                ext_data_type = DataType.normalize(extension_parts[0])
                if not data_type.is_super_type_of(ext_data_type):
                    return False
            return True
    else:
        # noinspection PyUnusedLocal
        def _predicate(extension: Extension) -> bool:
            return True

    return _predicate


#######################################################
# Classes
#######################################################


class DataOpener(ABC):
    """
    An interface that specifies a parameterized `open_data()` operation.

    Possible open parameters are implementation-specific and
    are described by a JSON Schema.

    Note this interface uses the term "opener" to underline the expected
    laziness of the operation. For example, when a xarray.Dataset is
    returned from a Zarr directory, the actual data is represented by
    Dask arrays and will be loaded only on-demand.
    """

    @abstractmethod
    def get_open_data_params_schema(self, data_id: str = None) \
            -> JsonObjectSchema:
        """
        Get the schema for the parameters passed as *open_params* to
        :meth:open_data(data_id, open_params).
        If *data_id* is given, the returned schema will be tailored
        to the constraints implied by the identified data resource.
        Some openers might not support this, therefore *data_id*
        is optional, and if it is omitted, the returned schema will be
        less restrictive.

        :param data_id: An optional data resource identifier.
        :return: The schema for the parameters in *open_params*.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def open_data(self, data_id: str, **open_params) -> Any:
        """
        Open the data resource given by the data resource identifier
        *data_id* using the supplied *open_params*.

        Raises if *data_id* does not exist.

        :param data_id: The data resource identifier.
        :param open_params: Opener-specific parameters.
        :return: An xarray.Dataset instance.
        :raise DataStoreError: If an error occurs.
        """


class DataDeleter(ABC):
    """
    An interface that specifies a parameterized `delete_data()` operation.

    Possible delete parameters are implementation-specific and
    are described by a JSON Schema.
    """

    @abstractmethod
    def get_delete_data_params_schema(self, data_id: str = None) \
            -> JsonObjectSchema:
        """
        Get the schema for the parameters passed as *delete_params*
        to :meth:delete_data.
        If *data_id* is given, the returned schema will be tailored to
        the constraints implied by the identified data resource.
        Some deleters might not support this, therefore *data_id*
        is optional, and if it is omitted, the returned schema will
        be less restrictive.

        :param data_id: An optional data resource identifier.
        :return: The schema for the parameters in *delete_params*.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def delete_data(self, data_id: str, **delete_params):
        """
        Delete a data resource. Raises if *data_id* does not exist.

        :param data_id: A data resource identifier known to exist.
        :param delete_params: Deleter-specific parameters.
        :raise DataStoreError: If an error occurs.
        """


class DataWriter(DataDeleter, ABC):
    """
    An interface that specifies a parameterized `write_data()` operation.

    Possible write parameters are implementation-specific and
    are described by a JSON Schema.
    """

    @abstractmethod
    def get_write_data_params_schema(self) -> JsonObjectSchema:
        """
        Get the schema for the parameters passed as *write_params* to
        :meth:write_data(data resource, data_id, open_params).

        :return: The schema for the parameters in *write_params*.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def write_data(self,
                   data: Any,
                   data_id: str,
                   replace: bool = False,
                   **write_params) -> str:
        """
        Write a data resource using the supplied *data_id* and *write_params*.

        :param data: The data resource's in-memory representation
            to be written.
        :param data_id: A unique data resource identifier.
        :param replace: Whether to replace an existing data resource.
        :param write_params: Writer-specific parameters.
        :return: The data resource identifier used to write the data resource.
        :raise DataStoreError: If an error occurs.
        """


class DataTimeSliceUpdater(DataWriter, ABC):
    """
    An interface that specifies writing of time slice data.
    """

    @abstractmethod
    def append_data_time_slice(self,
                               data_id: str,
                               time_slice: xr.Dataset):
        """
        Append a time slice to the identified data resource.

        :param data_id: The data resource identifier.
        :param time_slice: The time slice data to be inserted.
            Must be compatible with the data resource.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def insert_data_time_slice(self,
                               data_id: str,
                               time_slice: Any,
                               time_index: int):
        """
        Insert a time slice into the identified data resource at given index.

        :param data_id: The data resource identifier.
        :param time_slice: The time slice data to be inserted.
            Must be compatible with the data resource.
        :param time_index: The time index.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def replace_data_time_slice(self,
                                data_id: str,
                                time_slice: Any,
                                time_index: int):
        """
        Replace a time slice in the identified data resource at given index.

        :param data_id: The data resource identifier.
        :param time_slice: The time slice data to be inserted.
            Must be compatible with the data resource.
        :param time_index: The time index.
        :raise DataStoreError: If an error occurs.
        """
